#!/usr/local/bin/perl
#--------------------------------------------------------------------------
#
# dbamon_mssql.pl
#
#--------------------------------------------------------------------------
# Changes:
#      08Jun02 bb - Added event_limit code
# 1.10 25Oct02 bb - Better diag code.
# 1.11 14Jan04 bb - Recoded to not batch SQL statements
# 1.12 24Jan04 bb - Now, PERLPATH will always contain ../perl.exe
# 1.13 07Mar04 bb - New code to detect rsh service
# 1.14 19Apr04 bb - Fixed pslist code to end on error invoking 
# 1.15 14May04 bb - New NTRESKIT code. Now, try to find it in the default path
# 1.16 13Aug04 bb - New checkTLogs logic.
# 1.17 19Aug04 bb - New code for TLogs to check disk utilization
# 1.18 24Nov04 bb - Line 576 - added sort to the foreach
# 1.19 20Dcc04 bb - Added code to change color of WWW display if agent is down 
# 1.20 29Apr05 bb - New logic to query MSSQL version from db @@version 
# 1.21 11Aug05 bb - For MSSQL version, pad right node if < 4 chars (8.00.761 becomes 8.00.0761)
# 1.22 26Feb07 bb - New doubletake logic 
# 1.23 17Jul07 bb - Added dbatools to srvinfo search logic 
# 1.24 16Oct07 bb - For TEMPDB, we will now skip sp_spaceused and select max refdate
# 1.25 12Dec07 bb - New no_sp_spaceused up parm to skip sp_spaceused 
# 1.26 18Dec07 bb - Gather DB attributes (like snapshot, readonly)
# 1.27 07Jan07 bb - oldPerl - Changed hours threshold from 12 to 24 
# 1.28 16Jan07 bb - Suppress errors from cluster and srvinfo. Also recorded connect
#                     logic and new connectDB subroutine. Also, added isolation level
# 1.29 29Jan07 bb - Implement isolation level change for all servers 
# 2.00 05Jan11 bb - Massive changes.
# 2.10 08Feb11 bb - More Massive changes. New PERF collection and have to rethink TEMP
# 2.11 24Mar11 bb - Added saveConfig 
# 2.12 24Apr11 bb - diskCheck now uses WMI 
# 2.13 01Aug11 bb - New BADPARM checks in getDBAttrs 
# 2.14 15Sep11 bb - L572 - If we hist the SQL error log limit, then FINISH
# 2.15 29Sep11 bb - New code to deal with a SQL 2000 NAMED INSTANCE! (yuck) 
# 2.16 03Oct11 bb - New checkSuspectPages
# 2.17 05Oct11 bb - New LOCAL CFG concept (readCfg) 
# 2.18 11Nov11 bb - New io_delay_events logic AND new BULK SYSPERF
# 2.19 22Nov11 bb - WMI change - call CMD if WMI no-workie
# 2.20 22Nov11 bb - New countCPUs sub
# 2.21 10Dec11 bb - Improved timestamp checking in checklog
# 2.22 10Jan12 bb - New SysPerfSpecial for SQL 2000 
#--------------------------------------------------------------------------
$version = "2.22";
$| = 1;
$P = "dbamon_mssql.pl";

#--------------------------------------------------------------------------
# Variables
#--------------------------------------------------------------------------

use Win32::Registry;
use Time::Local;
use DBI;

#---- Get rid of compile warnings
$bt_desc = $tsts = $jobsupdated = $fg_fileid = $fg_growth = $char_val = $size = $usr2 = $root_fs = $hoursup = $endc = $pgm2 = $fg_file_in_group = $index_val = ""; 
$bt_dbname = $bt_type = $bt_timestamp = $bt_size  = ""; 
$connected_to_db = 0; 

($pshort = $P) =~ s/.pl$//; 
#TEMP#$plog =  "\\temp\\${pshort}_PgmLog.txt"; 
$plog =  "${pshort}_PgmLog.txt"; 
open(PLOG,">$plog") || &error("Cannot open output plog=$plog - $!"); 

$userid = Win32::LoginName();
$osname = Win32::GetOSName();
$in_instance = $ENV{"MSSQL_INSTANCE"}; 
$in_instance =~ s/ //g;
if (! $in_instance)
{
  &msg("Instance not specified - Using default MSSQLSERVER instance name"); 
  $in_instance = "MSSQLSERVER"; 
  $default_instance = 1;
}
else
{
  $default_instance = 0;
}

$domain = Win32::DomainName();
chop($host=`hostname`); 
$host_uc = uc $host; 

$debug = 0;
#$# = 2;
$event_limit = 100;
$fs = "<font size=2>"; 

#------------------------------------------------------------------------------
#
# Subroutines
#
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
sub commify 
#------------------------------------------------------------------------------
{
  # commify a number. Perl Cookbook, 2.17, p. 64
  my $text = reverse $_[0];
  $text =~ s/(\d\d\d)(?=\d)(?!\d*\.)/$1,/g;
  return scalar reverse $text;
}

#------------------------------------------------------------------------------
sub timeToNum
#------------------------------------------------------------------------------
{
  $zz_text = $_[0];
  $zz_text =~ s/-/\//g;
  $zz_text =~ s/:/\//g;
  ##print "zz_text=$zz_text\n";
  @words = split(/\//,$zz_text);

  if ($#words != 5)
  {
    &error("Input Target_End_Time ($zz_text) should be in the format: YYYY/MM/DD-HH:MM:SS","DOC116");
  }

  $zz_yyyy = $words[0];
  $zz_mon  = $words[1];
  $zz_dd   = $words[2];
  $zz_hh   = $words[3];
  $zz_mm   = $words[4];
  $zz_ss   = $words[5];
  ##print "zz_yyyy=$zz_yyyy zz_mon=$zz_mon zz_dd=$zz_dd zz_hh=$zz_hh zz_mm=$zz_mm zz_ss=$zz_ss\n";

  &error("Input Target_End_Time - YYYY must be 2002-2099","DOC116") if ($zz_yyyy < 2002 || $zz_yyyy > 2099);
  &error("Input Target_End_Time - MON  must be 1-12     ","DOC116") if ($zz_mon  <    1 || $zz_mon  >   12);
  &error("Input Target_End_Time - DD   must be 1-31     ","DOC116") if ($zz_dd   <    1 || $zz_dd   >   31);
  &error("Input Target_End_Time - HH   must be 0-23     ","DOC116") if ($zz_hh   <    0 || $zz_hh   >   23);
  &error("Input Target_End_Time - MM   must be 0-59     ","DOC116") if ($zz_mm   <    0 || $zz_mm   >   59);
  &error("Input Target_End_Time - SS   must be 0-59     ","DOC116") if ($zz_ss   <    0 || $zz_ss   >   59);

  @zz_date=($zz_ss,$zz_mm,$zz_hh,$zz_dd,--$zz_mon,$zz_yyyy);
  $zz_result = &timelocal(@zz_date);

  return $zz_result;
}

#------------------------------------------------------------------------------
sub getDBAttrs
#------------------------------------------------------------------------------
{
  #---- 1.26 START - Gather DB attributes (2005+ only)
  if ($ver >= 9)
  {
    &msg("DEBUG: This is an MSSQL 2005+ DB - Checking for DB attributes") if $debug;
    $sql = "
    select
      source_database_id, 
      is_read_only ,
      compatibility_level,
      recovery_model_desc,
      database_id,
      tempdb.dbo.dateconvert_19(create_date),
      state_desc
    from sys.databases where name = '$dbname';
    ";
    &msg("DEBUG: Running sql=$sql") if $debug;

    $sth = $dbh->prepare("$sql") || &error("Prepare");
    $sth->execute || &error("Execute_Select");
    $db_attrs = ""; 
    while (@Data = $sth->fetchrow_array)
    { 
      $source_database_id = $Data[0];
      $is_read_only       = $Data[1];
      $comp_level         = $Data[2];
      $rec_model          = $Data[3];
      $database_id        = $Data[4];
      $cre_date           = $Data[5];
      $db_state           = $Data[6];
      &msg("DEBUG: source_database_id=$source_database_id is_read_only=$is_read_only") if $debug;
    }

    $db_attrs .= "DBState=$db_state,CompatibilityLevel=$comp_level,CreateDate=$cre_date,DBID=$database_id,RecoveryModel=$rec_model";

    if ($source_database_id > 0) { $db_attrs .= "Snapshot,"; }
    if ($is_read_only == 1)      { $db_attrs .= "Readonly,"; }

    $dbname_out = $dbname;
    $dbname_out =~ s/\s/[BLANK]/g;
    $dbname_out =~ s/\-/[HYPHEN]/g;

    if ($db_attrs)
    {
      ##chop $db_attrs;    #---- Remove trailing comma
      print "$P DB_ATTRS: $dbname_out $db_attrs\n"; 
    }
  }
  else
  {
    &msg("DEBUG: This is NOT a MSSQL 2005+ DB - Not checking DB attributes");
  }
  #---- 1.26 END

  if (! $WWW)
  {
    #---- 2.13 START
    $sql = "SELECT 
    'autoshrink', 
    DATABASEPROPERTYEX ( '$dbname' , 'isautoshrink' ) 
    UNION ALL
    select 
    'autoclose',  
    DATABASEPROPERTYEX ( '$dbname' , 'isautoclose' ) 
    order by 1
    "; 
    $sth = $dbh->prepare($sql) || &dbError("Prepare_L3546");
    $sth->execute || &dbErrorOK("execute helpdb db=$dbname");

    $badparms = ""; 
    while (@Data = $sth->fetchrow_array)
    {
      $colname = $Data[0]; 
      $value   = $Data[1]; 
      ##print "DEBUG - BADPARMS-Select - $colname=$value\n";
      if ($value == 1) 
      { 
	$z = uc $colname;
	$badparms .= "$z=$value,"; 
      }
    }
    &msg("BADPARMS Results - badparms=($badparms)");
    if ($badparms)
    {
      $badparms =~ s/\,$//;  #---- Remove trailing comma
      $dbname_out = $dbname;
      $dbname_out =~ s/\s/[BLANK]/g;
      $dbname_out =~ s/\-/[HYPHEN]/g;
      print "$P BADPARMS: $dbname_out $badparms\n"; 
    }
    #---- 2.13 END
  }
}

#------------------------------------------------------------------------------
sub get_date
#------------------------------------------------------------------------------
{
  # Return date in format supplied as argument. Valid formats are:
  #   Y/M/D-H:M:S
  #   YMD_HMS
  #   YMD
  $format = $_[0];

  @sysdate = localtime;
  for ($i = 0; $i <= 5; $i++)
  {
    if ($sysdate[$i] < 10)
    {
      $sysdate[$i] = 0 . $sysdate[$i];
    }
  }

  if ($format eq "Y/M/D-H:M:S")
  {
    $sysdate = $sysdate[5]+1900 . "/" . ++$sysdate[4] . "/" . $sysdate[3] . "-" . $sysdate[2] . ":" . $sysdate[1] . ":" . $sysdate[0];
  }
  elsif ($format eq "YMD_HMS")
  {
    $sysdate = $sysdate[5]+1900 . ++$sysdate[4] . $sysdate[3] . "_" . $sysdate[2] . $sysdate[1] . $sysdate[0];
  }
  elsif ($format eq "YMD")
  {
    $sysdate = $sysdate[5]+1900 . ++$sysdate[4] . $sysdate[3];
  }
  else
  {
    &error("Invalid format passed to get_date. Must be Y/M/D-H:M:S or YMD_HMS");
  }
  return $sysdate;
}

#------------------------------------------------------------------------------
sub msg
#------------------------------------------------------------------------------
{
  $msgmsg = $_[0];
  $date = &get_date("Y/M/D-H:M:S");
  print "$date ${P} $version | $msgmsg\n";
  #---- Also log to a file
  print PLOG "$date ${P} $version | $msgmsg\n";
}

#------------------------------------------------------------------------------
sub dbError 
#------------------------------------------------------------------------------
{
   local($errorMsg,@error_fields) = @_;
   print "======== D B   E r r o r   E n c o u n t e r e d =========\n";
   print "Error: DB Error Encountered While: $errorMsg\n";
   print "Error: DB Error String: $DBI::errstr\n";
   print "==========================================================\n";
   $dberror = 1;
   @errors = split ("Database",$DBI::errstr);
   $chk_str = lc($errors[1]);
   $db      = $dbname;
   if ($chk_str =~ " \'$db\' cannot be opened because it is offline")
     {
       $offline = 1;
     }
   &msg("END_OF_PROBE");
   exit 1;
}

#------------------------------------------------------------------------------
sub dbErrorOK 
#------------------------------------------------------------------------------
{
   local($errorMsg,@error_fields) = @_;
   print "======== D B   E r r o r   E n c o u n t e r e d =========\n";
   print "Error: DB Error Encountered While: $errorMsg\n";
   print "Error: DB Error String: $DBI::errstr\n";
   print "==========================================================\n";
   $dberror = 1;
   @errors = split ("Database",$DBI::errstr);
   $chk_str = lc($errors[1]);
   $db      = $dbname;
   if ($chk_str =~ " \'$db\' cannot be opened because it is offline")
     {
       $offline = 1;
     }
}

# ----------------------------------------------------------------------
sub error
# ----------------------------------------------------------------------
{
  $msg = $_[0];
  print "$P | Fatal Error: $msg\n";
  print "***********************************************************\n";
  print "Error: $msg\n";
  print "***********************************************************\n";
  &msg("<font color=red><b>ERROR: $msg</b></font>");
  print "$P | Exiting\n";
  &msg("END_OF_PROBE");
  close(PLOG); 
  exit 1;
}

# ----------------------------------------------------------------------
sub bkbl
# ----------------------------------------------------------------------
{
  $dbname = $arg1;
  $bktype = $arg2;

  $dbname =~ s/\[HYPHEN\]/\-/g;
  $dbname =~ s/\[BLANK\]//g;
  
  if ($dbname eq "ALL")
  {
    if (-e "$SQLPATH\\dba_bkup_log\\ALL.txt")
    {
      $logpath = "$SQLPATH\\dba_bkup_log\\ALL.txt";
    }
    else
    {
      &msg("Looking for ALL.log in alternate location");
      $logpath = "\\temp\\ALL.txt";
    }
  }
  else
  {
    $logpath = "$SQLPATH\\dba_bkup_log\\${dbname}_${bktype}_backlog.txt";
  }

  &msg("bkbl1() - dbname=$dbname bktype=$bktype logpath=$logpath");

  if (! -e $logpath)
  {
    &error("Log file $logpath does not exist");
  }

  print "<table bgcolor=skyblue class=c1>
  <tr><td><pre>
<b>Start of mssql_backup Log: $logpath</b><hr>";

# - added by Scott.  I added the \" around the $logpath variable to accept spaces
# - in the file name and/or path.
  system("type \"$logpath\" 2>&1");

  print "<hr>
<b>End of mssql_backup Log: $logpath</b></table>\n"; }

# ----------------------------------------------------------------------
sub checklog
# ----------------------------------------------------------------------
{
  if ($arg1 > 1 && $arg1 < 30) { $minsev = $arg1; }
  else                         { $minsev = 17; }

  #---- Vars
  $cl_debug1 = 0; 
  $out_limit = 50; 
  my $io_delay_events = 0; 
  my $mem_paged_out_events = 0; 
  my $use_bookmark_file = 0; 

  &msg("checklog - Starting - event_limit=$event_limit - minsev=$minsev");  
  
  &msg("The following strings will be excluded (\@cfg_chkalert_e):");  
  foreach (@cfg_chkalert_e)
  {
    &msg("  $_");
  }

  $tsfile = "dbamon_mssql_checklog_ts_$in_instance.txt";
  $events = 0;
  
  if ($use_bookmark_file)
  {
    #---- OLD - FROM File
    &msg("(TSFILE-Start) Reading ts file: $tsfile");
    ## system("dir $tsfile");
    ## system("type $tsfile");
    #---- Read TS file
    open(TS, "$tsfile") || print "$P | Can't open input $tsfile - $!\n";;
    while (<TS>)
    {
      @words = split;
      $tsdate = $words[0];
      $tstime = $words[1];
    }
    $in_rtimestamp = "$tsdate $tstime"; 
    close(TS);
    &msg("(TSFILE-End) Input timestamps - We will begin processing at: tsdate=$tsdate tstime=$tstime - in_rtimestamp=$in_rtimestamp");
  }

  #---- NEW - From DCF
  $in_rtimestamp = $dcf_data{'SQLLogBookmark'}; 
  &msg("(DCF) Input dcf_data{SQLLogBookmark}=$dcf_data{'SQLLogBookmark'} - We will begin processing at in_rtimestamp=$in_rtimestamp");
  &msg("(DCF) PHONEHOME: Input DCF SQLLogBookMark is NULL") if ! $dcf_data{'SQLLogBookmark'};
  &msg("(DCF) PHONEHOME: Input in_rtimestamp is NULL") if ! $in_rtimestamp;

  #---- Common logic
  &msg("We will begin processing at in_rtimestamp=$in_rtimestamp - minsev=$minsev"); 

  use DBI;
  &connectDB;

  if (!$dberror)
  {
    &msg("Invoking master:xp_readerrorlog");
    $dbh->do("use master")                       || &dbError("Use_Master_1");
    $sth = $dbh->prepare("exec xp_readerrorlog") || &dbError("Prepare");

    $rows_raw = 0; 
    $rows_new = 0; 

    my $found_ts_good = 0; 
    my $found_ts_bad  = 0; 
    my $first_timestamp = ""; 
    my $first_timestamp_filtered = ""; 

    $sth->execute                                || &dbError("Execute_Select");
    READLOG: while (@Data = $sth->fetchrow_array)
    {
      $rows_raw++; 

      if ($vershort <= 80) 
      {
        @words = split ' ', $Data[0];
        $rdate = $words[0];
        $rtime = $words[1];
        $rpid  = $words[2];
        $verb  = $words[3];
      }
      else
      {
        @words = split ' ', $Data[0];
        $rdate = $words[0];
        $rtime = $words[1];
        @words = split ' ', $Data[1];
        $rpid  = $words[0];
        @words = split ' ', $Data[2];
        $verb  = $words[0];
      }

      #---- NEW way
      #---- Better Date validation
      #---- Example: Most Recent Date/Time 2011-12-10/05:45:09.23
      if (length($rdate) == 10 && $rdate =~ /^2/ && substr($rdate,4,1) eq "-" 
          && substr($rtime,2,1) eq ":" && substr($rtime,5,1) eq ":")
      {
        #---- Good timestamp 
        $found_ts_good++; 
        $save_rdate = $rdate; 
        $save_rtime = $rtime;
        $first_timestamp = "$save_rdate/$save_rtime" if ! $first_timestamp;
        $rtimestamp = "$rdate $rtime"; 
      }
      else
      {
        #---- Bad timestamp 
        $found_ts_bad++; 
        #---- We only want to keep rows with a good timestamp, which will exclude 2000 CONTINUATION rows
        next;
      }

      #---- OLD way
      #if ($rdate =~ /^20/)
      #{
      #  $save_rdate = $rdate;
      #  $save_rtime = $rtime;
      #}
     
      print "DEBUGLOG: TS-Compare - rtimestamp=$rtimestamp in_rtimestamp=$in_rtimestamp\n" if $cl_debug1;

      #---- Skip rows with a timestamp lower than our minimum timestamp
      if ($rtimestamp le $in_rtimestamp)
      {
        next; 
      }

      $first_timestamp_filtered = "$save_rdate/$save_rtime" if ! $first_timestamp_filtered;
      $rows_new++; 

      #---- Count messages that are not in the proper format
      if ($Data[2] =~ /requests taking longer than/i) { $io_delay_events++; } 
      if ($Data[2] =~ /significant part of sql server process memory has been paged out/i) { $mem_paged_out_events++; } 
  
      print "DEBUGLOG: HIT\n" if $cl_debug1; 

      if ($firstflag != 1)
      {
        &msg("First Date/Time $rdate/$rtime");
        $firstflag = 1;
      }

      &msg("Found New SQL Log Message - verb=$verb rdate=$rdate rtime=$rtime flag2nd=$flag2nd [Data0=$Data[0]] [Data1=$Data[1]] [Data2=$Data[2]]") if $cl_debug1;

      if ($flag2nd)
      {
        if ($vershort <= 80) 
        {
          $errmsg = substr($Data[0],32);
        }
        else
        {
          $errmsg = substr($Data[2],0);
        }
        $errmsg =~ s/\n$//;

        if ($errsev >= $minsev)
        {
          #---- OLD Way
          #---- Events that we don't want to report
          ## next if $errnum eq "17805";
          #---- NEW Way
          #---- See if this errnum is excluded
	  foreach $this_exclusion (@cfg_chkalert_e)
	  {
            if ($this_exclusion eq $errnum)
            {
              &msg("** This errnum=$errnum DOES match CFG File this_exclusion=$this_exclusion - NOT Reporting - Nexting to next log row");
              next READLOG;
            }
	  }

          #---- Now that we know that this row was NOT EXCLUDED, increment counter
          ++$events;
          &msg("Total Non-EXCLUDED Event Count (so far): $events");

          #---- Report this event - This is what we look for at the MASTER
          print "\n";
          print "=========================================================================\n";
          print "Event: $rdate-$rtime Error=$errnum Sev=$errsev Sta=$errsta Text=$errmsg\n";
          print "=========================================================================\n";
          print "\n";
        }
        $flag2nd = 0;
        next; 
      }

      if ($verb eq 'Error:')
      {
        &msg("Found SQL Log ERROR Message - verb=$verb rdate=$rdate rtime=$rtime flag2nd=$flag2nd [$Data[0]]");

        if ($events > $event_limit)
        {
          &msg("Event count of $events exceeds limit of $event_limit - will not process an more events");
          last;
        }

	if ($vershort <= 80) 
	{
	  @words = split ' ', $Data[0];
	  $errnum = $words[4]; 
	  chop($errnum);
	  $errsev = $words[6];
	  chop($errsev);
	  $errsta = $words[8];
	  chop($errsta);
	}
	else
	{
	  @words = split ' ', $Data[2];
	  $errnum = $words[1]; 
	  chop($errnum);
	  $errsev = $words[3];
	  chop($errsev);
	  $errsta = $words[5];
	  chop($errsta);
        }

        $flag2nd = 1; # We want to read the next line
        &msg("  Results - errnum=$errnum errsev=$errsev errsta=$errsta flag2nd=$flag2nd [$Data[2]]");
      }
    }  #---- End of fetch while

    if ($use_bookmark_file)
    {
      #OLD - #---- Save the last timestamp that we checked
      &msg("Saving timestamps: $save_rdate $save_rtime to $tsfile");
      open(TS, "> $tsfile")    || die "Error: Can't open output file: $tsfile\n";
      print TS "$save_rdate $save_rtime\n";
      close(TS);
      &msg("Contents of tsfile=$tsfile:");
      system("type $tsfile"); 
    }

    #---- Example: Most Recent Date/Time 2011-12-10/05:45:09.23
    &msg("Results from xp_readerrorlog:");
    &msg("  TotalRows:                $rows_raw");
    &msg("  NewRows:                  $rows_new");
    &msg("  ErrorsWithSev >= $minsev:      $events");
    &msg("  found_ts_good:            $found_ts_good"); 
    &msg("  found_ts_bad:             $found_ts_bad");
    &msg("  Log Low Timestamp:        $first_timestamp");
    &msg("  Bookmark Start Timestamp: $in_rtimestamp");
    &msg("  Filtered Low Timestamp:   $first_timestamp_filtered");
    &msg("  Log High Timestamp:       $save_rdate/$save_rtime");

    #---- Trigger event if > 1m rows
    print "$P SQL_Log_Rows: $rows_raw\n" if $rows_raw > 1000000; 

    #---- NEW - Save the last timestamp that we checked
    $dcf_data{'SQLLogBookmark'} = "$save_rdate $save_rtime";
  }

  #---- Just incase we had to "last" out of that while
  $sth->finish || &dbErrorOK("Finish_572");

  &msg("Special events:");
  #---- Event text if we found io_delay_events
  &msg("  I/O Delay Events: $io_delay_events");
  if ($io_delay_events > 0) 
  {
    print "$P IODelayEvents: $io_delay_events\n"; 
  }

  #---- Event text if we found mem_paged_out_events
  &msg("  Memory Paged Out Events: $mem_paged_out_events");
  if ($mem_paged_out_events > 0) 
  {
    print "$P MemPagedOutEvents: $mem_paged_out_events\n"; 
  }

  #---- Extract the last 30 lines of the SQL log for the drilldown page

  $dbh->do("use master")                       || &dbError("Use_Master_1");
  $sth = $dbh->prepare("exec xp_readerrorlog") || &dbError("Prepare");
  $sth->execute                                || &dbError("Execute_Select");

  undef %clog_hash; 
  $row_cnt   = 0; 
  $out_cnt   = 0; 
  &msg("Reading the last $out_limit lines of xp_errorlog for display at the MASTER"); 

  while (@Data = $sth->fetchrow_array)
  {
    $row_cnt++;  #---- For uniqueness in keynum

    if ($vershort <= 80)  #---- SQL 2000 & lower (less than 2005)
    {
      $keynum = substr($Data[0],0,23) . $row_cnt; 
      ##$zline = "<td><b> $Data[0] </b></td> "; 
      @words = split ' ', $Data[0];
      $ts1 = shift @words; 
      $ts2 = shift @words; 
      $spid = shift @words; 
      $text = ""; 
      foreach (@words) { $text .= "$_ "; }
      $zline = "
	<td nowrap> &nbsp; $ts1 $ts2 &nbsp; </td>
	<td nowrap> &nbsp; $spid &nbsp; </td>
	<td><b> $text </b></td>
      "; 
    }
    else   #---- > SQL 2000
    {
      $keynum = $Data[0] . $row_cnt; 
      $zline = "
	<td nowrap> &nbsp; $Data[0] &nbsp; </td>
	<td nowrap> &nbsp; $Data[1] &nbsp; </td>
	<td><b> $Data[2] </b></td>
      "; 
    }

    $keynum =~ s/\-//g;
    $keynum =~ s/\ //g;
    $keynum =~ s/\://g;
    $keynum =~ s/\.//g;

    ##&msg("DEBUG - keynum=$keynum"); 
    $zline =~ s/\n//g; 
    $clog_hash{$keynum} = $zline; 
  }

  foreach $key (sort hashDescTS (keys(%clog_hash))) 
  {
    last if $out_cnt > $out_limit;
    print "$P CurrentLog: " . $clog_hash{$key} . "\n"; 
    $out_cnt++; 
  }
  &msg("checklog - Done"); 
}

# ----------------------------------------------------------------------
sub checkAgentLog
# ----------------------------------------------------------------------
{
  # Vars
  $cl_debug1 = 0; 
  $out_limit = 50; 
  my $use_bookmark_file = 0; 

  &msg("checkAgentLog - Starting - event_limit=$event_limit");  
  
  #&msg("The following strings will be excluded (\@cfg_chkalert_e):");  
  #foreach (@cfg_chkalert_e)
  #{
  #  &msg("  $_");
  #}

  $events = 0;

  if ($use_bookmark_file)
  {
    #---- Old
    $tsfile = "dbamon_mssql_checkAgentLog_ts_${in_instance}.txt";
    &msg("Reading ts file: $tsfile");
    system("dir $tsfile");
    system("type $tsfile");
    #---- Read TS file
    open(TS, "$tsfile") || print "$P | Can't open input $tsfile - $!\n";;
    while (<TS>)
    {
      $ts = $_;
    }
    $in_rtimestamp = $ts; 
    &msg("Input timestamps - We will begin processing at: in_rtimestamp=$in_rtimestamp");
    close(TS);
  }

  if ($vershort < 90)
  {
    $sql_agent_log = "${SQLPATH}\\log\\sqlagent.out"; 
    if (-e "$sql_agent_log")
    {
      $sql = "exec xp_readerrorlog 1,'$sql_agent_log'"; 
    }
    else
    {
      &msg("SQL agent log path is non-standard - Agent log check cannot continue"); 
      &msg("  sql_agent_log=$sql_agent_log"); 
      return;
    }
  }
  else
  {
    $sql = "exec xp_readerrorlog 0,2"; 
  }
   
  #---- New - DCF
  $in_rtimestamp = $dcf_data{'AgentLogBookmark'}; 
  &msg("Input timestamps - DCF - dcf_data{'AgentLogBookmark'}=$dcf_data{'AgentLogBookmark'} - We will begin processing at in_rtimestamp=$in_rtimestamp");
  &msg("(DCF) PHONEHOME: Input DCF AgentLogBookMark is NULL") if ! $dcf_data{'AgentLogBookmark'};

  &msg("Invoking master:xp_readerrorlog - sql=$sql");
  $dbh->do("use master")                           || &dbError("Use_Master_1");
  $sth = $dbh->prepare($sql) || &dbError("Prepare");

  $rows_raw = 0; 
  $rows_new = 0; 

  $sth->execute                                || &dbError("Execute_Select");
  while (@Data = $sth->fetchrow_array)
  {
    $rows_raw++; 

    if ($vershort < 90) 
    {
      $ag_ts         = substr($Data[0],0,19);
      $ag_text       = substr($Data[0],21);
      $ag_errorlevel = "";  
    }
    else
    {
      $ag_ts         = $Data[0];
      $ag_errorlevel = $Data[1];
      $ag_text       = $Data[2];
    }

    print "DEBUGLOG-AG: TS-Compare - ag_ts=$ag_ts in_rtimestamp=$in_rtimestamp\n" if $cl_debug1;

    if ($ag_ts le $in_rtimestamp)
    {
      next; 
    }
    $rows_new++; 
    print "DEBUGLOG: HIT - rows_new=$rows_new\n" if $cl_debug1; 

    if ($firstflag != 1)
    {
      &msg("First ag_ts=$ag_ts");
      $firstflag = 1;
    }

    &msg("Found New SQL Log Message - ag_ts=$ag_ts ag_errorlevel=$ag_errorlevel ag_text=$ag_text") if $cl_debug1;

    #---- REPORT any errors
    if ($ag_text =~ /error/i)
    {
      ++$events;
      &msg("Total Event Count (so far): $events");

      #---- Events that we don't want to report
      next if $ag_text =~ /idunno/i;

      #---- Report this event 
      print "\n";
      print "=========================================================================\n";
      print "EventAgent: TS=$ag_ts ErrorLevel=$ag_errorlevel - $ag_text\n";
      print "=========================================================================\n";
      print "\n";
    }
  } 

  &msg("Most Recent ag_ts=$ag_ts");

  &msg("From xp_readerrorlog - TotalRows: $rows_raw NewRows: $rows_new Errors: $events - ag_ts=$ag_ts");

  if ($use_bookmark_file)
  {
    #---- Save the last timestamp that we checked
    &msg("Saving timestamp: $ag_ts to $tsfile");
    open(TS, "> $tsfile")    || die "Error: Can't open output file: $tsfile\n";
    print TS "$ag_ts\n";
    close(TS);
    &msg("Contents of tsfile=$tsfile:");
    system("type $tsfile"); 
  }

  #----NEW - DCF #---- Save the last timestamp that we checked
  $dcf_data{'AgentLogBookmark'} = "$ag_ts";
  
  #---- Extract the last 30 lines of the SQL log for the drilldown page

  $dbh->do("use master")                           || &dbError("Use_Master_1");
  $sth = $dbh->prepare($sql) || &dbError("Prepare");
  $sth->execute                                    || &dbError("Execute_Select");

  undef %clog_hash; 
  $row_cnt   = 0; 
  $out_cnt   = 0; 
  &msg("Reading the last $out_limit lines of xp_errorlog for display at the MASTER"); 

  while (@Data = $sth->fetchrow_array)
  {
    $row_cnt++;  #---- For uniqueness in keynum

    $keynum = substr($Data[0],0,23) . $row_cnt;    #---- Timestamp
    $zline = "
      <td nowrap> &nbsp; $Data[0] &nbsp; </td>
      <td nowrap> &nbsp; $Data[1] &nbsp; </td>
      <td><b> $Data[2] </b></td>
      "; 

    $keynum =~ s/\-//g;
    $keynum =~ s/\ //g;
    $keynum =~ s/\://g;
    $keynum =~ s/\.//g;

    &msg("DEBUG - keynum=$keynum") if $cl_debug1; 
    $zline =~ s/\n//g; 
    $clog_hash{$keynum} = $zline; 
  }

  foreach $key (sort hashDescTS (keys(%clog_hash))) 
  {
    last if $out_cnt > $out_limit;
    print "$P CurrentLogAgent: " . $clog_hash{$key} . "\n"; 
    $out_cnt++; 
  }
  &msg("checkAgentLog - Done");  
}

# ----------------------------------------------------------------------
sub hashDescTS 
# ----------------------------------------------------------------------
{
  $b <=> $a
}

# ----------------------------------------------------------------------
sub checkTLog
# ----------------------------------------------------------------------
{
  &msg("Checking transaction logs for all databases");

  if ($WWW)
  {
    print "<br>\n";
    &msg("Transaction Log Status:");
    print "<br><table border=1 bgcolor=skyblue class=c1>\n";
    print "<tr bgcolor=slate>
      <th><h5>Database<br>(One Row Per TLOG)
      <th><h5>TLog Size<br>(mB)
      <th><h5>TLog<br>Percent<br>Full<br>Internal
      <th><h5>TLog<br>Percent<br>Full<br>vs. Max
      <th><h5>Is DB In<br>Trunc-Log-<br>On-Ckpt<br>Mode?
      <th><h5>TLog<br>Max<br>Size
      <th><h5>TLog<br>Path
      <th><h5>TLog<br>Drive
      <th><h5>TLog<br>Drive<br>Size<br>(Mb)
      <th><h5>TLog<br>Drive<br>Free<br>(Mb)
      <th><h5>TLog<br>Drive<br>Used<br>(Mb)
      <th><h5>TLog<br>Drive<br>Percent<br>Full
    \n";
  }

  $sql = "exec ('DBCC SQLPERF (LOGSPACE)')";

  $dbh->do("use master") || &dbError("Use_Master");
  $sth3 = $dbh->prepare($sql) || &error("Could not prepare sql=$sql");
  $sth3->execute() || &error("Could not execute sql=$sql");
  while (@Data = $sth3->fetchrow_array)
  {
    $selectcount++;

    $ctl_dbname  = $Data[0];
    $ctl_size{"$ctl_dbname"}    = sprintf("%.2f",$Data[1]);
    $ctl_pctused{"$ctl_dbname"} = sprintf("%.2f",$Data[2]);
  }
  $sth3->finish() || &dbError("Could not execute FINISH");

  ##@sorted = sort keys %ctl_size;
  foreach (sort keys %ctl_size)
  {
    $dbname = $_;   
    ##print "DEBUG_TLOG - dbname=$dbname\n";  
    $size    = $ctl_size{"$dbname"};
    $pctused = $ctl_pctused{"$dbname"};

    &dbProperties($dbname);

    $dbp_rawdata = "";
    &dbProperties2($dbname);
    foreach (split('\n',$dbp_rawdata))
    {
      @words = split;

      ##print "TLOG_DEBUG: $_\n"; 

      ##$dbp_rawdata .= "$dbp_log_fname $dbp_log_space $dbp_log_fsize\n";

      $dbp_log_fname = $words[0];
      $dbp_log_space = $words[1];
      $dbp_log_fsize = $words[2];

      #---- Start
      if ($dbp_trunc_on_ckpt == 1)
      {
	$trunc = "Yes";
      }
      else
      {
	$trunc = "No";
      }

      #---- See how full the TLog disk is
      $this_drive_letter = uc substr($dbp_log_fname,0,1);

      #---- Old
      #$this_drive_size   = (split(" ",$disks{"$this_drive_letter"})) [2];
      #$this_drive_free   = (split(" ",$disks{"$this_drive_letter"})) [3];
      #$this_drive_used   = (split(" ",$disks{"$this_drive_letter"})) [4];
      #$this_drive_pctused = sprintf("%.2f",($this_drive_used / $this_drive_size) * 100) if $this_drive_size > 0;

      #---- New
      $this_drive_size   = commify($disk_size_mb{"$this_drive_letter"}); 
      $this_drive_free   = commify($disk_free_mb{"$this_drive_letter"}); 
      $this_drive_used   = commify($disk_used_mb{"$this_drive_letter"}); 
      $this_drive_pctused = $disk_pct_full{"$this_drive_letter"}; 

      $dbname_out = $dbname;
      $dbname_out =~ s/\s/[BLANK]/g;
      $dbname_out =~ s/\-/[HYPHEN]/g;

      if ($WWW)
      {
	if    ($trunc eq "Yes") { $color = "skyblue"; }
	elsif ($pctused >= 90 || $this_drive_pctused >= 90)  { $color = "pink"; }
	elsif ($pctused >= 80 || $this_drive_pctused >= 80)  { $color = "yellow"; }
	else                    { $color = "lime"; }

        if ($dbp_log_space ne "Unlimited")
        {
          $pctused_vs_max = sprintf "%.2f", ($dbp_log_fsize / $dbp_log_space) * 100;
	  if    ($pctused_vs_max >= 99) { $color = "pink"; }
	  elsif ($pctused_vs_max >= 95) { $color = "yellow"; }
        }
        else
        {
          $pctused_vs_max = "N/A";
        }
	
	print "<tr bgcolor=$color>
	  <td><h5> <b> $dbname_out </b>
	  <td align=right><h5> $dbp_log_fsize
	  <td align=right><h5> ${pctused}%
	  <td align=right><h5> ${pctused_vs_max}%
	  <td><h5> $trunc
	  <td><h5> $dbp_log_space
	  <td><h5> $dbp_log_fname
	  <td><h5> $this_drive_letter
	  <td align=right><h5> $this_drive_size
	  <td align=right><h5> $this_drive_free
	  <td align=right><h5> $this_drive_used
	  <td align=right><h5> ${this_drive_pctused}%
	\n";
      }
      else
      {
        $this_drive_pctused = "." if $this_drive_pctused eq ""; 
        if ($dbp_log_space == 2147483.65) { $dbp_log_space = "Unlimited"; }
        if ($dbp_log_space eq "2147483.65") { $dbp_log_space = "Unlimited"; }

	print "$P TLOG: $dbname_out $trunc $dbp_log_fsize $pctused $this_drive_letter $this_drive_pctused $dbp_log_fname $dbp_log_space\n";
      } 

      #---- End
    }
  }

  if ($WWW)
  {
    print "</table><br>\n";
  }

  &msg("Found $selectcount databases");

  if ($WWW)
  {
    print "<br>\n";
    &msg("Drives Which Contain MSSQL Data Files:");
    print "<br><table border=1 bgcolor=skyblue class=c1>\n";
    print "<tr bgcolor=slate>
	<th><h5>Drive 
	<th><h5>Size (mB)
	<th><h5>Used (mB)
	<th><h5>Free (mB)
	<th><h5>Percent<br>Full
    ";
  }

  #---- TEST
  foreach (sort @db_drivelist)
  {
    $this_drive = $_;
    if ($_ ne $last)
    {
      #---- Old
      #$this_drive_size   = t(" ",$disk_size_gb{"$this_drive"})) [2];
      #$this_drive_free   = (split(" ",$disks{"$this_drive"})) [3];
      #$this_drive_used   = (split(" ",$disks{"$this_drive"})) [4];
      #$this_drive_pctused = sprintf("%.2f",($this_drive_used / $this_drive_size) * 100) if $this_drive_size > 0;
      #---- New
      $this_drive_size   = commify($disk_size_mb{"$this_drive"}); 
      $this_drive_free   = commify($disk_free_mb{"$this_drive"}); 
      $this_drive_used   = commify($disk_used_mb{"$this_drive"});  
      $this_drive_pctused = $disk_pct_full{"$this_drive"}; 

      if    ($this_drive_pctused >= 90) { $color = "pink"; }
      elsif ($this_drive_pctused >= 80) { $color = "yellow"; }
      else                              { $color = "lime"; }

      if ($WWW)
      {
	print "<tr bgcolor=$color>
	    <td><h5> $this_drive
	    <td align=right><h5> $this_drive_size
	    <td align=right><h5> $this_drive_used
	    <td align=right><h5> $this_drive_free
	    <td align=right><h5> ${this_drive_pctused}%
	";
      }
      else
      {
	print "$P DRIVESTATUS: $this_drive $this_drive_size $this_drive_used $this_drive_free $this_drive_pctused \n";
      }
    }
    $last = $_;
  }

  if ($WWW)
  {
    print "</table><br>\n";

    &msg("WWWTEST");
    #&test;
  }

  return 0;
}

# ----------------------------------------------------------------------
sub dbProperties2
# ----------------------------------------------------------------------
{
   my $dbname = $_[0];

   #&msg("dbProperties2 - executing sp_helpdb \@dbname='$dbname'"); 

   $sth = $dbh->prepare("exec sp_helpdb \@dbname='$dbname'") || &dbError("Prepare");
   $sth->execute || &dbErrorOK("sp_helpdb db=$dbname");

   $dbp_log_space = "";

   do
   {
     ##print "BEFORE WHILE\n";
     while(@Data = $sth->fetchrow_array)
     {
       #$i = -1;
       #foreach (@Data)
       #{
       #  $i++;
       #  $z = $_;
       #  print "Data[$i]=$z\n"; 
       #}
       
       if ($Data[7] eq "log only") 
       { 
         $dbp_log_fname  = $Data[2]; 
         $dbp_log_fsize  = $Data[4]; 
         $dbp_log_space  = $Data[5]; 
         $dbp_log_growth = $Data[6]; 

	 $dbp_log_fname =~ s/ //g;
	 $dbp_log_fsize =~ s/ //g;
	 $dbp_log_space =~ s/ //g;

         if ($dbp_log_space =~ /2147483648/) 
         { 
           if ($dbp_log_growth eq "0 KB") { $dbp_log_space = "FixedSize"; }
           else                           { $dbp_log_space = "Unlimited"; }
         }

         ##print "TLOG_DEBUG: dbname=$dbname dbp_log_space=$dbp_log_soace dbp_log_growth=$dbp_log_growth\n"; 

	 if ($dbp_log_space =~ /KB$/)
	 {
	   chop $dbp_log_space;
	   chop $dbp_log_space;
	   $dbp_log_space = sprintf "%.2f", ($dbp_log_space * 1024) / (1024*1024);
	 }
	 if ($dbp_log_fsize =~ /KB$/)
	 {
	   chop $dbp_log_fsize;
	   chop $dbp_log_fsize;
	   $dbp_log_fsize = sprintf "%.2f", ($dbp_log_fsize * 1024) / (1024*1024);
	 }

         $dbp_rawdata .= "$dbp_log_fname $dbp_log_space $dbp_log_fsize\n";
         ##print "<tr bgcolor=white><td colspan=99>dbp_log_fname=$dbp_log_fname dbp_log_space=$dbp_log_space dbp_log_fsize=$dbp_log_fsize\n";
       }
     }
   } 
   while($sth->{odbc_more_results} && $dbp_log_space eq "");

   $sth->finish || &dbErrorOK("Finish1");

   return;
}


# ----------------------------------------------------------------------
sub dbFiles
# ----------------------------------------------------------------------
{
   my $dbname = $_[0];

   push(@db_drivelist,"C");

   $sth = $dbh->prepare("exec sp_helpdb \@dbname='$dbname'") || &dbError("Prepare");
   $sth->execute || &dbErrorOK("execute helpdb db=$dbname");

   my $db_filelist = "";

   do
   {
     while(@Data = $sth->fetchrow_array)
     {
       #$i = -1;
       #foreach (@Data)
       #{
       #  $i++;
       #  $z = $_;
       #  print "Data[$i]=$z\n"; 
       #}
       
       if ($Data[7] eq "data only") 
       { 
         $db_filelist .= " $Data[2]"; 
         my $db_filesize =  $Data[4]; 

	 my $db_filedrive = substr($Data[2],0,1);
         push(@db_drivelist,uc $db_filedrive);

         $db_filesize =~ s/ KB//;
         $db_filesize_mb = sprintf("%.2f", ($db_filesize / 1024)); 
         $db_tot_datafile_size_mb{$dbname} += $db_filesize_mb; 
         $db_tot_datafile_count{$dbname}++; 
       }
       if ($Data[7] eq "log only") 
       { 
         my $log_filesize =  $Data[4]; 

	 my $db_filedrive = substr($Data[2],0,1);
         push(@db_drivelist,uc $db_filedrive);

         $log_filesize =~ s/ KB//;
         $log_filesize_mb = sprintf("%.2f", ($log_filesize / 1024)); 
         $db_tot_log_size_mb{$dbname} += $log_filesize_mb; 
       }

     }
   } 
   while($sth->{odbc_more_results});
   $sth->finish || &dbErrorOK("Finish1");

   $tlog_db_ratio{$dbname} = sprintf("%.2f", ($db_tot_log_size_mb{$dbname} / $db_tot_datafile_size_mb{$dbname})) if $db_tot_datafile_size_mb{$dbname} > 0; 

   return $db_filelist;
}


# ----------------------------------------------------------------------
sub dbProperties
# ----------------------------------------------------------------------
{
   my $dbname = $_[0];

   $dbp_trunc_on_ckpt = -999;

   ##&msg("  Checking logging option for: $dbname...");

   my $sth4 = $dbh->prepare(<<End_SQL) or die "COULD NOT PREPARE SQL";
   select databaseproperty('$dbname','IsTruncLog') as log_option
End_SQL
   $sth4->execute() || &dbError("COULD NOT EXECUTE SQL STATEMENT");

   while (($log_option) = $sth4->fetchrow_array() )
   {
      $dbp_trunc_on_ckpt = $log_option;
   }
   $sth4->finish() || &dbError("COULD NOT EXECUTE SQL STATEMENT");

   return;
}

# ----------------------------------------------------------------------
sub getNetInfo
# ----------------------------------------------------------------------
{
  &msg("getNetInfo - Starting - Querying net properties"); 

  #---- IP and VLAN
  open(NET, 'ipconfig|') || print "Error: Running net start";
  while(<NET>)
  {
    ##print;
    @words = split;

    if (/IPv4 Address/) 
    {
      $net_ip = $words[$#words]; 
    }
    if (/IP Address/) 
    {
      $net_ip = $words[$#words]; 
    }
  }

  #---- Extract VLAN 
  @net_vlan_ary = split (/\./,$net_ip); 
  $net_vlan = $net_vlan_ary[0] . "." . $net_vlan_ary[1] . "." . $net_vlan_ary[2]; 

  print "$P NET_IP:      $net_ip\n";
  print "$P NET_VLAN:    $net_vlan\n";
  print "$P NET_COMPANY: $company\n";
  &msg("getNetInfo - Done - net_ip=$net_ip net_vlan=$net_vlan"); 
}

# ----------------------------------------------------------------------
sub up
# ----------------------------------------------------------------------
{
  #---- Test START
  ## $arg2 = "no_sp_spaceused" if $host eq "wcosdba2"; 
  #---- Test END
  &msg("up() Starting - arg1=$arg1 arg2=$arg2"); 
  
  #---- Get Network stuff
  &getNetInfo;

  #---- See if this is a PHYSICAL or VIRTUAL server (presence of c:\program files\vmware folder)
  if ( -e "c:\\program files\\vmware") { $server_type = "VirtualVMWare"; }
  else                                 { $server_type = "Physical"; }
  print "ServerType: $server_type\n";

  #---- See if this file exists (this logic is for the lightbulb)
  #TEMP#$sp_file = "C:\\temp\\DBAmon_no_sp_spaceused.txt";
  $sp_file = "DBAmon_no_sp_spaceused.txt";
  $sp_hours = 24;
  if (-e "$sp_file")
  {
    system("dir $sp_file"); 
    $sp_file_mtime = ( stat $sp_file )[9];
    $sp_file_age_hrs = (time - $sp_file_mtime) / 3600;
    &msg("up() sp_file=$sp_file Does exist - sp_file_age_hrs=$sp_file_age_hrs sp_file_mtime=$sp_file_mtime"); 
    if ($sp_file_age_hrs <= $sp_hours)
    {
      $arg2 = "no_sp_spaceused"; 
      &msg("  Since sp_file has been updated within the last $sp_hours hours, I will set arg2=$arg2"); 
    }
    else
    {
      &msg("  Since sp_file has NOT been updated within the last $sp_hours hours, no action taken"); 
    }
  }

  &msg("Checking DB up and collecting space data - arg1=$arg1 arg2=$arg2"); 

  #---- Old
  ## $numcpus = $ENV{"NUMBER_OF_PROCESSORS"};
  #---- New
  &countCPUs;

  %values="";
  $Register = "Hardware\\Description\\System\\CentralProcessor\\0";
  $HKEY_LOCAL_MACHINE->Open($Register,$hkey)|| die "Cannot open registry $Register - $!";
  $hkey->GetValues(\%values);
  foreach $value (keys(%values))
  {
    $RegType      = $values{$value}->[1];
    $RegValue     = $values{$value}->[2];
    $RegKey       = $values{$value}->[0];
    $MHz = $RegValue if $RegKey eq "\~MHz";
    $vendor = $RegValue if $RegKey eq "VendorIdentifier";
    $identifier = $RegValue if $RegKey eq "Identifier";
  }
  $hkey->Close();
  @cpu_ident = split (' ',$identifier); 
  $cpumodel = ${cpu_ident[0]} . "/" . ${cpu_ident[2]};
  $cpu = $vendor . " " . " " . $MHz . " MHz ";
 

  %values="";
  $Register = "Software\\Microsoft\\Windows NT\\CurrentVersion";
  $HKEY_LOCAL_MACHINE->Open($Register,$hkey)|| die "Cannot open registry $Register - $!";
  $hkey->GetValues(\%values);
  foreach $value (keys(%values))
  {
    $RegType      = $values{$value}->[1];
    $RegValue     = $values{$value}->[2];
    $RegKey       = $values{$value}->[0];
    $CurrentVersion = $RegValue if $RegKey eq "CurrentVersion";
    $CSDVersion = $RegValue if $RegKey eq "CSDVersion";
  }

  if ($CSDVersion ne "")
  {
    @sp_info = split(' ',$CSDVersion);
    $osver = $CurrentVersion . ",SP" . ${sp_info[2]};
  }
  else
  {
    $osver = $CurrentVersion;
  }

  $rsh_service = "?";

  # See if the MSSQLServer is running
  &msg("Running 'net start' to view running services - Looking for MSSQL running for this instance"); 
  open(NET, 'net start|') || print "Error: Running net start";
  while(<NET>)
  {
    ##print;
    print if /sql/i;
    @words = split;

    if ($rsh_service eq "?")
    {
      if (/Remote Shell Daemon/)  { $rsh_service="RSD"; }
      if (/Remote Shell Service/) { $rsh_service="RSS"; }
      if (/CYGWIN sshd/)          { $rsh_service="CYGWIN"; }
      if (/openssh/i)             { $rsh_service="OpenSSH"; }
    }

    if (lc($words[0]) eq "mssqlserver") { print; $servicestat="[Running]"; }
  
    #---- Old
    #if ($vershort < 90)
    #{
    #  $service_mode = "DefaultInstance";  
    #  if (/SQL Server \(/)                 { print; $servicestat="[Running]"; }
    #}
    #else
    #{
    #  $service_mode = "SpecificInstance";  
    #  if (/SQL Server \(/ && /$in_instance/) { print; $servicestat="[Running]"; }
    #}
    #---- New - may need more work for 2005
    if ($vershort < 90)
    {
      if ($default_instance)
      {
        if (/^MSSQLSERVER/) { print; $servicestat="[Running]"; }
      }
      else
      {
        if (/MSSQL\$$in_instance/i) { print; $servicestat="[Running]"; }
      }
    }
    else
    {
      if (/SQL Server \(/ && /$in_instance/) { print; $servicestat="[Running]"; }
    }
  }
  close(NET);
  &msg("Read through running services - default_instance=$default_instance - servicestat=$servicestat"); 

  #---- Get MSSQL license key 
  $mssql_regkey = "?"; 
  if    ($vershort == 80) { $swver_regkey = "Software\\Microsoft\\Microsoft SQL Server\\80\\Registration"; }
  elsif ($vershort == 70) { $swver_regkey = "Software\\Microsoft\\Microsoft SQL Server 7.0\\7.00.000\\Product ID"; }
  else  { $swver_regkey = ""; }
  if ($swver_regkey)
  {
    %values="";
    $Register = "Software\\Microsoft\\Microsoft SQL Server\\$vershort\\Registration";
    &msg("Looking for LK - swver_regkey=$swver_regkey"); 
    $HKEY_LOCAL_MACHINE->Open($swver_regkey,$hkey) || print "Uh-Oh1\n"; 
    if ($hkey)
    {
      $hkey->GetValues(\%values) || print "Uh-Oh\n";
      foreach $value (keys(%values))
      {
	$RegKey        = $values{$value}->[0];
	$RegType       = $values{$value}->[1];
	$RegValue      = $values{$value}->[2];
	##print "DEBUG RegKey=$RegKey RegValue=$RegValue\n"; 
	$mssql_regkey  = $RegValue if $RegKey eq "CD_KEY";
      }
      $hkey->Close();
    }
  }
  print "MSSQLSWKEY: $mssql_regkey\n";

  if ($servicestat eq "")
  {
    $servicestat="[Not Running]";
  } 

  if (!$WWW)
  {
    print "OSVer: $osver $osname\n";
    print "SQLVer: $ver\n";

    print "MSSQLServiceStatus: $servicestat\n";
    print "NumCPUs: $numcpus\n";
    print "CPU: $cpu\n";
    print "CPUModel: $cpumodel\n";
    print "HOSTNAME: $host\n";
    print "RSHService: $rsh_service\n";
  }

  if ($servicestat =~ /not running/i)
  {
    &msg("Since SERVICE for this MSSQL instance ($in_instance) is NOT RUNNING, I will not try to connect to the DB"); 
    print "DBStatus: DOWN\n";
    &msg("END_OF_PROBE");
    exit 0;
  }
   
  &connectDB if ! $dbh;

  #---- Get SysPerf data here... 
  if (! $WWW) 
  {
    $perf_newway = 1; 
    if ($perf_newway)
    {
      &SysPerfBulk; 
      &SysPerfSpecial; 
    }
    else
    {
      &SysPerf; 
      &SysPerfBulk; 
      &SysPerfCompare; 
      &SysPerfSpecial; 
    }
  }

  #---- Get memory Info
  $dbh->do("use master")                || &dbError("Use_Master");
  $mem = $dbh->prepare("exec xp_msver") || &dbError("Prepare");
  $mem->execute                         || &dbError("Execute_Select");

  while (($index_val,$name,$internal_val,$char_val) = $mem->fetchrow_array() )
  {
     if (lc($name) eq "physicalmemory")
     {
        $ram = $internal_val;
        last;
     }
  }
  $mem->finish() || &dbError("Could not execute FINISH");

  if (defined($ram) && !$WWW)
  {
     print "RAM: $ram MB\n";
  }

  #---- Get the SQL server version, patch level and edition
  &msg("Querying SQL for patchlevel and edition"); 

  $dbh->do("use master")                || &dbError("Use_Master");
  $sth = $dbh->prepare("
    SELECT SERVERPROPERTY('productversion'), 
    SERVERPROPERTY ('productlevel'), 
    SERVERPROPERTY ('edition'), 
    SERVERPROPERTY ('licensetype'), 
    SERVERPROPERTY ('numlicenses') ") || &dbError("Prepare_SQLVER_L1077");
  $sth->execute                         || &dbError("Execute_Select");

  $sql_version = $sql_patchlevel = $sql_edition = $sql_licensetype = $sql_numlicenses = "?"; 

  while (@Data = $sth->fetchrow_array)
  {
    $sql_version      = $Data[0]; 
    $sql_patchlevel   = $Data[1]; 
    $sql_edition      = $Data[2]; 
    $sql_licensetype  = $Data[3]; 
    $sql_numlicenses  = $Data[4]; 
  }
  $sth->finish() || &dbError("Could not execute FINISH");
  &msg("Results - sql_version=$sql_version sql_patchlevel=$sql_patchlevel sql_edition=$sql_edition sql_licensetype=$sql_licensetype sql_numlicenses=$sql_numlicenses"); 

  $sql_edition =~ s/\s//g if $sql_edition;
  $sql_numlicenses = -1 if ! $sql_numlicenses; 

  if (defined($sql_version) && defined($sql_patchlevel) && defined($sql_edition) && !$WWW)
  {
     print "$P SQL_Details: $sql_version $sql_patchlevel $sql_edition $sql_licensetype $sql_numlicenses\n";
  }

  #---- See if the agent is running

  &msg("Running 'net start' to see if the Agent is running");
  open(NET, 'net start|') || print "Error: Running net start";
  $agentrunning = 0; 
  while(<NET>)
  {
    print if /SQL/ && /agent/i;
    @words = split;

    if ($vershort < 90)  #---- SQL 2000 & lower (less than 2005)
    {
      if ($default_instance)
      {
	if (lc($words[0]) eq "sqlserveragent") { $agentrunning = 1; last; }
	if (lc($words[0]) eq "sqlexecutive") { $agentrunning = 1; last; }
      }
      else
      {
	if ($words[0] =~ /sqlagent\$${in_instance}/i) { $agentrunning = 1; last; }
      }
    }
    else  #---- 2005 or higher -- will always have the instance name as part of the service name
    {
      if (/SQL Server Agent \($in_instance\)/) { $agentrunning = 1; last; }
    }
  }
  close(NET);
  &msg("Resulting agentrunning=$agentrunning"); 

  # Attempt to connect to the master database on the server, to verify MSSQL availability.
  &connectDB;

  if (!$dberror)
  {
    &msg("Checking database=master");
    $dbh->do("use master") || &dbError("Use_Master");

    # Read from master
    $sql = "select name from sysdatabases order by 1; ";
    $sth = $dbh->prepare("$sql") || &dbError("Prepare");
    $sth->execute                || &dbError("Execute_Select");
    while (@Data = $sth->fetchrow_array)
    {
      $dbname = $Data[0];
      push(@dblist,$dbname);
      ++$numdbs;
    }

    &msg("DEBUG - selecting current_user") if $debug; 
    # Get session userid
    $sql = "select current_user; ";
    $sth = $dbh->prepare("$sql") 
      || &dbError("Prepare");
    $sth->execute
      || &dbError("Execute_Select");
    while (@Data = $sth->fetchrow_array)
    {
      $user = $Data[0];
    }
    ##print "Current_User: $user\n";

    # Process DB names
    foreach $dbname (@dblist)
    {
      ##$dbname_out = lc $dbname;  #---- 2010/11/25 - Let's try to leave it mixed case
      $dbname_out = $dbname;
      $dbname_out =~ s/\s/[BLANK]/g;
      $dbname_out =~ s/\-/[HYPHEN]/g;

      # Get info for this DB   
      &getDBAttrs;

      $sth->finish()        || &dbError("Finish");
      $dbh->do("use \"$dbname\"")        || &dbError("Use_$dbname");
      # Query space 

      ##&msg("DEBUG - Running exec sp_helpdb \@dbname='$dbname'"); 

      $sth = $dbh->prepare("exec sp_helpdb \@dbname='$dbname'") || &dbError("Prepare");
      $sth->execute || &dbErrorOK("sp_helpdb db=$dbname");

      $db_status = "";
      $db_created = "";
      $db_owner = "";

      while (@Data = $sth->fetchrow_array)
      {
        $db_owner   = $Data[2];
        $db_created = $Data[4];
        $db_status  = $Data[5];
        ##print "dbname=$dbname db_owner=$db_owner db_status=$db_status db_created=$db_created\n";   # Test
      }

      $db_spacealloc = "?";
      $db_spacefree  = "?";
      $db_spaceused  = "?";

      if ($arg2 eq "no_sp_spaceused")
      {
        &msg("Skipping sp_spaceused for dbname=$dbname since Probe_Parm: = no_sp_spaceused"); 
        #TEMP#$sp_file = "C:\\temp\\DBAmon_no_sp_spaceused.txt";
        $sp_file = "DBAmon_no_sp_spaceused.txt";
        &msg("Touching file $sp_file for other utilities"); 
        system("echo x > $sp_file"); 
      }
      elsif ($dbname eq "tempdb")
      {
        &msg("Skipping sp_spaceused for tempdb (hostname/DB $host/$dbname)"); 
      }
      else
      {
	&msg("Running exec sp_spaceused for dbname='$dbname'"); 
	$sth = $dbh->prepare("exec sp_spaceused") || &dbError("Prepare");
	$dberror = 0;
	$sth->execute || &dbErrorOK("sp_spaceused db=$dbname");
	if ($dberror)
	{
	  ++$numdbsoffline;
	  print "DB:     $dbname_out CONNECTERROR \n";
	  next;
	}
	while (@Data = $sth->fetchrow_array)
	{
	  ##print "Data0=$Data[0]\n";   # Test
	  ##@words = split ' ', $Data[0];
	  ##$verb = $words[0];
	  ##if ($verb eq $dbname)
	  if (uc $Data[0] eq uc $dbname)
	  {
	    $db_spacealloc = $Data[1];
	    chop $db_spacealloc ; chop $db_spacealloc ;chop $db_spacealloc ;
	    $db_spacefree  = $Data[2];
	    chop $db_spacefree ; chop $db_spacefree ;chop $db_spacefree ;
	    $db_spaceused  = $db_spacealloc - $db_spacefree;
	  }
	}
        $db_spacealloc = sprintf("%.2f", $db_spacealloc);
        $db_spaceused  = sprintf("%.2f", $db_spaceused );
      }

      $db      = $dbname;

      #---- Check last-accessed date
      $db_lastrefdate = "?"; 
      if ($dbname eq "tempdb")
      {
        &msg("Skipping select max refdate for tempdb (hostname/DB $host/$dbname)"); 
      }
      else
      {
	&msg("  Running 'select max(refdate) from sysobjects (nolock)' for dbname='$dbname'"); 
	$sth = $dbh->prepare("select max(refdate) from sysobjects (nolock)") || &dbError("Prepare");
	$sth->execute || &dbErrorOK("select sysobjects to db=$dbname");
	$db_lastrefdate = "?";
	while (@Data = $sth->fetchrow_array)
	{
	  $db_lastrefdate = $Data[0]; 
	}
      }

      $db_lastrefdate = "? ?" if $db_lastrefdate eq "?"; 
      $db_spacealloc = 0 if ! $db_spacealloc ; 
      $db_spaceused  = 0 if ! $db_spaceused  ; 
      $db_spaceused  = "?" if ! $db_spaceused  ; 

      print "DB:     $dbname_out $db_spacealloc $db_spaceused $db_lastrefdate $db_owner\n" if !$WWW;

      $dbfilelist = &dbFiles($dbname);

      push @www_dbs,"<td><h5>$dbname_out
                     <td nowrap align=right><h5> &nbsp; $db_spacealloc &nbsp; </td>
                     <td nowrap align=right><h5> &nbsp; $db_spaceused &nbsp; </td>
                     <td nowrap align=right><h5> &nbsp; $db_tot_datafile_size_mb{$dbname} &nbsp; </td>
                     <td nowrap align=right><h5> &nbsp; $db_tot_log_size_mb{$dbname} &nbsp; </td>
                     <td nowrap align=right><h5> &nbsp; $tlog_db_ratio{$dbname} &nbsp; </td>
                     <td nowrap><h5> &nbsp; $db_created &nbsp; </td>
                     <td nowrap><h5> &nbsp; $db_lastrefdate &nbsp; </td>
                     <td nowrap><h5> &nbsp; $db_owner &nbsp; </td>
                     <td ><h5> &nbsp; $dbfilelist &nbsp; </td>
                     <td nowrap><h5> &nbsp; $db_status &nbsp; </td>
      ";

      $tot_spacealloc += $db_spacealloc if $db_spacealloc ne "?";
      $tot_spaceused  += $db_spaceused  if $db_spaceused  ne "?";
    }

    if (!$WWW)
    {
      print "NUMDBs: $numdbs\n";
      print "SpaceAlloc: $tot_spacealloc\n";
      print "SpaceUsed: $tot_spaceused\n";
      &msg("DB Datafile Counts:"); 
      foreach (sort keys %db_tot_datafile_count)
      {
        print "$P DBDFCount: $_ $db_tot_datafile_count{$_}\n"; 
      }
    }
 
  }

  #---- Take 1st sp_monitor sample
  #---- 2011/02/10 BB - We don't need this any more since i now use typeperf
  ## &runSPMonitor;

  # Decide status
  if ($numdbs > 0) 
  {
    $num_up = $numdbs - $numdbsoffline;
    print "DBStatus: UP \($num_up\)" if !$WWW;
    if ($offline)
     {
       print "; DOWN \($numdbsoffline\)" if !$WWW;
     }
    print "\n" if !$WWW;
  }
  else
  {
    print "DBStatus: DOWN\n if !$WWW";
  }

  if ($agentrunning) 
  {
    print "AgentStatus: UP\n" if !$WWW;
    $www_agent_status = "Up";
  }
  else
  {
    print "AgentStatus: DOWN\n" if !$WWW;
    $www_agent_status = "Down";
  }

  #---- Old way
  ##if ($ENV{"NTRESKIT"}) { $srvinfo_cmd = '%NTRESKIT%\srvinfo -ns'; }
  ##else                  { $srvinfo_cmd = 'srvinfo -ns'; }
  ##&msg("Running srvinfo - srvinfo_cmd=$srvinfo_cmd");

  #---- New way
  #---- See if srvinfo is in the system path
  #if (-e "C:\\dbatools\\srvinfo_dbamon.exe") 
  #{
  #  $srvinfo_cmd = "C:\\dbatools\\srvinfo_dbamon.exe";
  #}
  #---- 2011/04 BB - This was moved to &init
  #if (system("srvinfo -ns >NUL 2>NUL") == 0) { $srvinfo_cmd = "srvinfo -ns"; }
  #elsif ($ENV{'NTRESKIT'}) { $srvinfo_cmd = "$ENV{'NTRESKIT'}\\srvinfo -ns"; }
  #elsif (-e "dbamon_mssql_di.exe") { $srvinfo_cmd = "dbamon_mssql_di.exe -ns"; }
  #else
  #{
  #  &msg("Hmmmm... Could not find srvinfo in default path and NTRESKIT not in env");
  #}

  #---- 2011/04 BB - This was obsoleted by driveCheck
  #&msg("Running $srvinfo_cmd (A)");
  #open(SRVINFO, "$srvinfo_cmd 2>&1 |") || &error("Error: Running $srvinfo_cmd");
  #$uptime = "?";
  #while(<SRVINFO>)
  #{
  #  ##print;
  #  @words = split;
  #
  #    $os64bit = 1 if /EM64/;
  #
  #    if (/^System Up Time:/)
  #    {
  #      chop($uptime = substr($_,16));
  #    }
  #
  #    #---- 1.17 - Save this for later
  #    #if ($#words == 4 && ($words[1] eq "FAT" || $words[1] eq "NTFS"))
  #    if (length($words[0]) == 2 && ($words[1] eq "FAT" || $words[1] eq "NTFS"))
  #    {
  #      $drive = substr($words[0],0,1);
  #      $disks{"$drive"} = $_;
  #      $disk_size_gb{"$drive"} = sprintf("%.2f",$words[2] / 1024);
  #      $disk_free_gb{"$drive"} = sprintf("%.2f",$words[3] / 1024);
  #      $disk_used_gb{"$drive"} = sprintf("%.2f", ($disk_size_gb{"$drive"} - $disk_free_gb{"$drive"}));
  #      $disk_size_mb{"$drive"} = sprintf("%.2f",$words[2] );
  #      $disk_free_mb{"$drive"} = sprintf("%.2f",$words[3] );
  #      $disk_used_mb{"$drive"} = sprintf("%.2f", ($disk_size_mb{"$drive"} - $disk_free_mb{"$drive"}));
  #      $disk_pct_full{"$drive"} = sprintf("%.2f", 100 * ($disk_used_gb{"$drive"} / $disk_size_gb{"$drive"})) if  $disk_size_gb{"$drive"} > 0;
  #    }
  #  }
  #  &msg("Ran $srvinfo_cmd to capture uptime - uptime=$uptime");

  #---- Collect info about local drives
  &driveCheck; 
  &msg("Disk info:");
  foreach $drive (sort keys %disks)
  {
    &msg("  Disk: $drive - Size=$disk_size_gb{$drive} Free=$disk_free_gb{$drive} Used=$disk_used_gb{$drive}  PctF=$disk_pct_full{$drive}"); 
  }

  #---- 2011/04/08 - We only want to run SYSTEMINFO if we did not collect $mem_avail in SysPerf - it takes way too long
  #---- 2011/04/08 -   Note that i don't care about system_hardware_info - it isn't passed back to MASTER 
  if ($mem_avail eq "") 
  {
    #---- Systeminfo START - gather MEM usage info
    ##if ($osver >= 5.2 && $host =~ /wcosplm/)
    # System Boot Time:          11/25/2010, 7:57:04 AM
    # System Manufacturer:       Hewlett-Packard
    # System Model:              HP Compaq 6530b (NA407UC#ABA)
    # System Type:               x64-based PC
    if ($osver >= 5.2 && -e "$ENV{'windir'}\\system32\\systeminfo.exe")
    {
      &msg("Running systeminfo to gather memory usage information"); 
      $system_hardware_info = ""; 
      open(SYSINFO, "systeminfo 2>&1 |") || &msg("Error: Running systeminfo - $!");
      while(<SYSINFO>)
      {
	##print;
	@words = split;
	if (/^Available Physical Memory:/)
	{
	  $mem_avail_raw = $words[$#words - 1] . $words[$#words];
	  $mem_avail = $words[$#words - 1] if $words[$#words] eq "MB";
	  $mem_avail =~ s/\,//g;
	} 
	if (/^System Boot Time:/) { $system_boot_time = $words[3] . $words[4] . $words[5]; }
	##if (/^System Manufacturer:/) { $system_hardware_info .= $words[2] . ":"; }
	if (/^System Model:/)     { $system_hardware_info .= "$words[2]_$words[3]_$words[4]_$words[5]";  }
	if (/^System Type:/)      { $system_hardware_info .= "_$words[2]_$words[3]";  }
      }
      close(SYSINFO); 

      
      &msg("DEBUG - mem_avail=$mem_avail mem_avail_raw=$mem_avail_raw ram=$ram mem_used=$mem_used mem_pct=$mem_pct"); 
      if ($mem_avail_raw ne "")
      {
	#TEMP#open(MEM, ">>c:\\temp\\MemAvail.txt") || &msg("Error: Cannot open output file: MemAvail.txt - $!");
	open(MEM, "MemAvail.txt") || &msg("Error: Cannot open output file: MemAvail.txt - $!");
	print MEM "$date $mem_avail_raw\n";
	close(MEM);
      }
    }
    #---- Systeminfo END - gather MEM usage info
  }

  #---- This could have come from SysPerf or from the above SYSTEMINFO
  if ($ram > 0 && $mem_avail ne "")
  {
    $mem_used = $ram - $mem_avail;
    $mem_pct  = sprintf("%.2f",($mem_used / $ram) * 100); 
  }

  #---- Cluster status - START
  $cluster_name = "";
  open(CLUSTERV, "cluster /ver 2>&1 |"); 
  while(<CLUSTERV>)
  {
    ##print;
    #Cluster Name:     LSCASQLC
    @words = split;
    $cluster_name = $words[2] if (/^Cluster Name:/)
  }
  close(CLUSTERV); 
  &msg("cluster /ver Shows cluster_name=$cluster_name"); 
  if ($cluster_name ne "") 
  {
    &msg("Querying cluster status - $cluster_name"); 
    open(CLUSTER, "cluster /cluster:$cluster_name res |") || &error("Error: Running cluster");
    $cluster_status = "";
    while(<CLUSTER>)
    {
      ##print;
      @words = split;
      if (/^SQL Server  /)
      {
        $cluster_status = $words[$#words - 0];
        $cluster_node   = $words[$#words - 1];
        if ($cluster_status ne "Online")
        {
          $cluster_thisnodestatus   = "Offline"; 
        }
        else
        {
	  if ($cluster_node eq uc $host) { $cluster_thisnodestatus   = "Active"; }
	  else                           { $cluster_thisnodestatus   = "Passive"; }
    
        }
        ##&msg("  cs=$cluster_status cn=$cluster_node"); 
        ##print;
      }
    }
    close(CLUSTER); 

    print "$P CLUSTERNAME: $cluster_name\n"; 
    print "$P CLUSTERSTATUS: $cluster_status\n"; 
    print "$P CLUSTERONNODE: $cluster_node\n"; 
    print "$P CLUSTERTHISNODESTATUS: $cluster_thisnodestatus\n"; 
  }
  #---- Cluster status - END

  &msg("Querying MSSQL for MSSQL software version");
  $sth->finish()        || &dbError("Finish2");
  $sth = $dbh->do("use master") || &dbError("Use_Master");
  $sql = "select \@\@version; ";
  $sth = $dbh->prepare("$sql") || &dbError("Prepare");
  $sth->execute || &dbError("Execute_Select");
  while (@Data = $sth->fetchrow_array)
  {
    $mssql_version_raw  = $Data[0];
    $mssql_version_from_db = (split(/ /,$mssql_version_raw))[6];
    if (length((split(/\./,$mssql_version_from_db)) [2]) == 3)
    {
      $mssql_version_from_db =
             (split(/\./,$mssql_version_from_db)) [0] . '.' .
	     (split(/\./,$mssql_version_from_db)) [1] . '.' .
	     "0" . (split(/\./,$mssql_version_from_db)) [2];
      &msg("DEBUG = Padding right node of version with 0 - mssql_version_from_db=$mssql_version_from_db"); 
    }
  }
  $sth->finish || &dbError("Execute_Finish");
  &msg("DEBUG ver=$ver mssql_version_from_db=$mssql_version_from_db");

  if ($mssql_version_from_db gt $ver && $mssql_version_from_db)
  {
    &msg("Overriding MSSQL version from db since it is higher - mssql_version_from_db=$mssql_version_from_db");
    $ver = $mssql_version_from_db; 
    print "SQLVer: $ver\n" if ! $WWW;
  }

  &msg("Querying MSSQL startup time");
  $sth->finish()        || &dbError("Finish2");
  $sth = $dbh->do("use master") || &dbError("Use_Master");
  $sql = "select 
      login_time,
      datediff(hour,login_time, getdate()) as hours_since
    from  
      sysprocesses where status = 'background' order by spid; ";
  $sth = $dbh->prepare("$sql") || &dbError("Prepare");
  $sth->execute || &dbError("Execute_Select");
  while (@Data = $sth->fetchrow_array)
  {
    $mssql_started  = $Data[0];
    $mssql_hours_up = $Data[1];
    last;
  }
  $sth->finish || &dbError("Execute_Finish");
  &msg("mssql_started=$mssql_started mssql_hours_up=$mssql_hours_up");

  $mssql_started_fix = substr($mssql_started,0,19);
  $mssql_started_fix =~ s/\-/\//g;
  $mssql_started_fix =~ s/ /-/g;
  if ($mssql_started_fix =~ /^1900/)
  {
    $mssql_started_fix =~ s/^1900/2002/;
  }

  &msg("mssql_started_fix=$mssql_started_fix");
  $str_c = $mssql_started_fix;
  if ($str_c)
  {
    $str_n = &timeToNum($str_c);
    $endc = $str_c;
    $calc_date = $date;
    $calc_time = time;

    if (! $WWW)
    {
      print "$P UPTIME:          $str_c $str_n $calc_date $calc_time\n";
      print "$P MSSQL64BIT: 1 \n" if $mssql64bit;
      print "$P OS64BIT: 1 \n"    if $os64bit;
      ##&msg("DEBUG - mem_avail=$mem_avail mem_avail_raw=$mem_avail_raw ram=$ram mem_used=$mem_used mem_pct=$mem_pct"); 
      print "$P RAM:       $ram\n"         if $ram;
      print "$P MEM_AVAIL: $mem_avail\n"   if $mem_avail;
      print "$P MEM_USED:  $mem_used\n"    if $mem_used;
      print "$P MEM_PCT:   $mem_pct\n"     if $mem_pct;
    }
  }

  #---- Calc CPU BUSY - We did this earlier also
  #---- 2011/11/14 BB - Let's not do this any more - arithmetic exceptions on servers that have been up for a long time
  ##&runSPMonitor; 
  ##print "$P CPU_BUSY_WITH_SQL: $cpu_busy\n"; 

  if ($WWW)
  {
    $www_bgcolor = "pink";
    $www_bgcolor = "lime" if $numdbs > 0;

    #---- Exceptions
    $www_bgcolor = "pink" if $www_agent_status ne "Up";

    $www_status = "Down";
    $www_status = "Up" if $numdbs > 0;

    $daysup = $hoursup / 24;
    $daysup = sprintf("%.2f", $daysup);

    $mssql64bit_out = "64bit" if $mssql64bit;
    $os64bit_out = "64bit"    if $os64bit;

    &msg("WWW Status Summary:");
    print " <style type='text/css'>
	table.c1 {font-family:consolas; font-size:10pt; }
      </style> "; 
    print "<table border=1 bgcolor=skyblue class=c1>\n";
    print "<tr bgcolor=slate><th><h5>Item<th><h5>Status\n";
    print "<tr bgcolor=$www_bgcolor><td><b>Hostname:</b>  <td>$host\n";
    print "<tr bgcolor=$www_bgcolor><td><h5>OS Version<td><h5>$osver $os64bit_out $osname\n";
    print "<tr bgcolor=$www_bgcolor><td><b>MSSQL Instance:</b><td>$in_instance\n";
    print "<tr bgcolor=$www_bgcolor><td><h5>MSSQL Status<td><h5>$www_status\n";
    print "<tr bgcolor=$www_bgcolor><td><h5>MSSQL Service Status<td><h5>$servicestat\n";
    print "<tr bgcolor=$www_bgcolor><td><h5>MSSQL Agent Status<td><h5>$www_agent_status\n";
    print "<tr bgcolor=$www_bgcolor><td><h5>MSSQL Start Date<td><h5>$mssql_started_fix\n";
    print "<tr bgcolor=$www_bgcolor><td><h5>MSSQL Hours Active<td><h5>$mssql_hours_up\n";
    print "<tr bgcolor=$www_bgcolor><td><h5>Database Count<td><h5>$numdbs\n";
    print "<tr bgcolor=$www_bgcolor><td><h5>Database Space Alloc.<td><h5>$tot_spacealloc mB\n";
    print "<tr bgcolor=$www_bgcolor><td><h5>Database Space Used<td><h5>$tot_spaceused mB\n";
    print "<tr bgcolor=$www_bgcolor><td><h5>MSSQL Version<td><h5>$ver $mssql64bit_out\n";
    print "<tr bgcolor=$www_bgcolor><td><h5>CPU Info<td><h5>$cpu $cpumodel\n";
    print "<tr bgcolor=$www_bgcolor><td><h5>CPU Count<td><h5>$numcpus\n";
    print "<tr bgcolor=$www_bgcolor><td><b>Server Info: </b> <td> $system_hardware_info \n";
    print "<tr bgcolor=$www_bgcolor><td><b>Server Last Reboot: </b> <td> $system_boot_time \n";
    print "<tr bgcolor=$www_bgcolor><td><h5>RAM Total<td><h5>$ram mB\n";
    print "<tr bgcolor=$www_bgcolor><td><h5>RAM Available<td><h5>$mem_avail mB\n";
    print "<tr bgcolor=$www_bgcolor><td><h5>RAM Used<td><h5>$mem_used mB\n";
    print "<tr bgcolor=$www_bgcolor><td><h5>RAM % Used<td><h5>${mem_pct}%\n";
    ##print "<tr bgcolor=$www_bgcolor><td><h5>Server Uptime<td><h5>$uptime\n";
    print "<tr bgcolor=$www_bgcolor><td><h5>Remote Shell Service<td><h5>$rsh_service\n";
    print "<tr bgcolor=$www_bgcolor><td><h5>MSSQL LK<td><h5>$mssql_regkey\n" if $mssql_regkey ne "?";
    print "<tr bgcolor=$www_bgcolor><td><h5>Cluster Name (If Any)<td><h5>$cluster_name\n" if $cluster_name;
    print "<tr bgcolor=$www_bgcolor><td><h5>Cluster Global Status (If Any)<td><h5>$cluster_status\n" if $cluster_status;
    print "<tr bgcolor=$www_bgcolor><td><h5>Cluster Active On Node (If Any)<td><h5>$cluster_node\n" if $cluster_node;
    print "<tr bgcolor=$www_bgcolor><td><h5>Cluster This-Node Status (If Any)<td><h5>$cluster_thisnodestatus\n" if $cluster_thisnodestatus;
    print "<tr bgcolor=$www_bgcolor><td><h5>DoubleTake Status<td><h5>$doubletake_status\n" if $doubletake_status;
    print "<tr bgcolor=$www_bgcolor><td><h5>DoubleTake Dir<td><h5>$doubletake_dir\n" if $doubletake_dir;
    print "<tr bgcolor=$www_bgcolor><td><h5>DoubleTake File<td><h5>$doubletake_file\n" if $doubletake_file;
    print "<tr bgcolor=$www_bgcolor><td><h5>DoubleTake FileAge Mins<td><h5>$doubletake_fileage_mins\n" if $doubletake_fileage_mins;
    print "</table><br>\n\n";
  
    &msg("Databases:");
    print "<table border=1 bgcolor=skyblue class=c1>\n";
    print "<tr bgcolor=slate>
      <th><h5>Database<br>Name
      <th><h5>Alloc.<br>(mB)<br>DB+TLog
      <th><h5>Used<br>(mB)<br>DB+TLog
      <th><h5>Total Size (mB) Of DB File(s) 
      <th><h5>Total Size (mB) Of TLog File(s) 
      <th><h5>TLog Size to <br>DB Size <br>Ratio
      <td><h5>Date<br>Created
      <td><h5>Most Recent Object <br>Created Date
      <td><h5>Owner
      <th><h5>Datafile(s)
      <td><h5>Status
    \n";
    foreach (@www_dbs)
    {
      ##print "<tr bgcolor=$www_bgcolor>$_ \n";
      print "<tr bgcolor=white>$_ \n";
    }
    print "<tr bgcolor=slate><td><b> $fs Total:
      <td align=right><b> $fs $tot_spacealloc
      <td align=right><b> $fs $tot_spaceused 
    </table>";
  
    &checkTLog;
    #&msg("TEST"); 
    #&bkup;

    &checkFilegroups;

    &msg("Done");
    &msg("END_OF_PROBE");
    exit 0;
  }
}

#------------------------------------------------------------------------------
sub runSPMonitor
#------------------------------------------------------------------------------
{
  &msg("runSPMonitor - Sampling CPU_BUSY from sp_monitor ");

  $sql = "exec sp_monitor";
  ##&msgl("Running DB=master sql=$sql");
  $sth = $dbh->prepare("$sql") || &error("Prepare");
  $sth->execute || &error("Execute_Select");
  while($href = $sth->fetchrow_hashref or $sth->{odbc_more_results})
  {
    # got what we needed, stop asking
    last if (defined $href->{cpu_busy});
  }
  $cpu_raw  = $href->{cpu_busy};
  $cpu_busy = (split(/\-/,$cpu_raw)) [1];
  $cpu_busy =~ s/\%//;
  &msg("runSPMonitor - Done - cpu_raw=$cpu_raw cpu_busy=$cpu_busy");
  $sth->finish || &error("Execute_FINISH");
}

#------------------------------------------------------------------------------
sub perfSample
#------------------------------------------------------------------------------
{
  $perfmode = $_[0];
  &msg("  perfSample - Starting - perfmode=$perfmode");

  #---- 1 Start. Total transaction count

  &msg("  Collecting TransRate sample"); 
  if ($vershort == 80) 
  {
    $sql = "select cntr_value from master.dbo.sysperfinfo 
      where counter_name = 'Transactions/sec' and instance_name = '_Total' "; 
  }
  else
  {
    $sql = "select cntr_value from master.sys.sysperfinfo 
      where counter_name = 'Transactions/sec' and instance_name = '_Total' "; 
  }

  $sth = $dbh->prepare("$sql") || &error("Prepare");
  $sth->execute || &error("Execute_Select");
  while(@Data = $sth->fetchrow_array)
  {
    $perf_all{"TransRate"} = $Data[0]; 
  }
  $sth->finish || &error("Execute_FINISH");

  #---- 1 End.   Total transaction count

  #---- 1 Start. Blocked sessions 
  $dbh->do("if OBJECT_ID('tempdb..#dbamon_spwho2') > 0 drop table #dbamon_spwho2") || &dbError("Drop_If_L1697");
  
  if ($vershort == 80) 
  {
    $dbh->do("create table #dbamon_spwho2 (
        SPID int not null
      , Status varchar (255) not null
      , Login varchar (255) not null
      , HostName varchar (255) not null
      , BlkBy varchar(10) not null
      , DBName varchar (255) null
      , Command varchar (255) not null
      , CPUTime int not null
      , DiskIO int not null
      , LastBatch varchar (255) not null
      , ProgramName varchar (255) null
      , SPID2 int not null
     )") || &dbError("Drop_If_L1697");
  }
  else
  {
    $dbh->do("create table #dbamon_spwho2 (
        SPID int not null
      , Status varchar (255) not null
      , Login varchar (255) not null
      , HostName varchar (255) not null
      , BlkBy varchar(10) not null
      , DBName varchar (255) null
      , Command varchar (255) not null
      , CPUTime int not null
      , DiskIO int not null
      , LastBatch varchar (255) not null
      , ProgramName varchar (255) null
      , SPID2 int not null
      , requestid int
     )") || &dbError("Drop_If_L1697");
  }
  $dbh->do("insert #dbamon_spwho2 exec sp_who2") || &dbError("Drop_If_L1697");
  
  $sql = "select COUNT(*)
    from #dbamon_spwho2
    where
    SPID > 50 and BlkBy != '  .  '
  "; 
  $sth = $dbh->prepare("$sql") || &error("Prepare");
  $sth->execute || &error("Execute_Select");
  while(@Data = $sth->fetchrow_array)
  {
    $perf_all{"BlockedSessions"} = $Data[0]; 
  }
  $sth->finish || &error("Execute_FINISH");
  
  $dbh->do("if OBJECT_ID('tempdb..#dbamon_spwho2') > 0 drop table #dbamon_spwho2") || &dbError("Drop_If_L1743");
 
  #---- 1 End.   Blocked sessions


  #---- Copy hash to the correct name
  if ($perfmode eq "Before") 
  { 
    %perf_before_all = %perf_all;
  }
  elsif ($perfmode eq "After") 
  {
    %perf_after_all = %perf_all;
  }
  else { &error("perfSample - Impossible perfmode=$perfmode") } 

  &msg("  perfSample - Done");
}

#------------------------------------------------------------------------------
sub SysPerfSample
#------------------------------------------------------------------------------
{
  $sps_counter = $_[0];
  $sps_label   = $_[1];
  &msg("  SysPerfSample - Starting - sps_counter=$sps_counter sps_label=$sps_label"); 
  $sps_file = "dbamon_typeperf_in.txt"; 
  $sps_val = "?"; 

  open(SPSFILE,">$sps_file") || &error("Cannot open output $sps_file - $!"); 
  print SPSFILE "$sps_counter\n"; 
  close(SPSFILE); 

  $sps_cmd = "typeperf.exe -cf $sps_file -sc 1"; 
  ##&msg("  Running sps_cmd=$sps_cmd"); 
  ##open(SPSCMD,"$sps_cmd|") || &error("Cannot run typeperf - $!"); 
  open(SPSCMD,"$sps_cmd|"); 
  while(<SPSCMD>)
  {
    next if /CSV/ || /^The command/ || /^Exiting/ || /^$/ || /^\s$/; 
    print;
    @sps_words = split (/\,/,$_) ; 
    if ($#sps_words == 1 && $sps_words[1] =~ /^\"/) 
    {
      $sps_val = $sps_words[1]; 
      $sps_val =~ s/\n//g;
    }
    last;
  }
  &msg("  Resulting sps_val=$sps_val"); 

  if ($sps_val eq "?") 
  {
    &msg("  Since not value was returned by TYPEPEF, returning prematurely"); 
    return;
  }

  $sps_val =~ s/\"//g; 
  $sps_val = sprintf("%.2f", $sps_val); 

  &msg("  SysPerfSample - Done - Displaying $sps_label $sps_val"); 
  print "$P SYSPERF: $sps_label $sps_val\n"; 
  $sysperf{$sps_label} = $sps_val; 

  &msg("  SysPerfSample - Done"); 
  return; 
}

#------------------------------------------------------------------------------
sub SysPerf
#------------------------------------------------------------------------------
{
  &msg("SysPerf - Starting");
  $sp_start = time; 

  #---- Sample
  #typeperf "\MSSQL$MSSQL1:Buffer Manager\Buffer cache hit ratio" -sc 1
  
  #---- Each DB seems to have a different way of specifying the first part of the verb 
  #---- So by default we will use "SQLServer:", but we will query sys.dm_os_performance_counters to get the real value
  $counter_prefix = "SQLServer";   #---- Default
  if ($vershort > 80)
  {
    &connectDB if ! $dbh; 
    &msg("Selecting from dm_os_performance_counters to obtain counter_prefix"); 
    $sql = "select top 1 OBJECT_NAME from sys.dm_os_performance_counters";
    $sth = $dbh->prepare("$sql") || &dbError("Prepare1954");
    $sth->execute                || &dbError("Execute_Select1955");
    while (@Data = $sth->fetchrow_array)
    {
      $counter_string = $Data[0];
    }
    $sth->finish || &dbErrorOK("Finish1960");
    $counter_prefix_raw = (split(":",$counter_string)) [0];
    $counter_prefix = $counter_prefix_raw if $counter_prefix_raw;
    &msg("Resulting counter_string=$counter_string counter_prefix_raw=$counter_prefix_raw"); 
  }
  &msg("Using counter_prefix=$counter_prefix"); 

  #---- Now collect the data from the windoze performance counters

  #---- NEW - Collect from parms in CFG file
  &msg("From CFG file: \%cfg_perfcounter"); 
  foreach (sort keys %cfg_perfcounter)
  {
    $counter     = $_; 
    $dbamon_verb = $cfg_perfcounter{$_};

    if ($counter =~ /SQLSERVERPREFIX/) { $counter =~ s/SQLSERVERPREFIX/$counter_prefix/; } 
    
    &msg("  Found: counter=$counter dbamon_verb=$dbamon_verb - Invoking SysPerfSample");
    &SysPerfSample($counter, $dbamon_verb); 
  }

  #---- OLD - Collector name and verbs HARD-CODED
  #&SysPerfSample("\\memory\\available mbytes", "MemAvailMBytes"); 
  #&SysPerfSample("\\system\\processor queue length", "CPUQueueLen"); 
  #&SysPerfSample("\\processor(_total)\\% processor time","CPUTotalBusyPct"); 
  #&SysPerfSample("\\${counter_prefix}:Buffer Manager\\Page life expectancy",   "SQL_PageLifeExpectancy"); 
  #&SysPerfSample("\\${counter_prefix}:Buffer Manager\\Buffer cache hit ratio", "SQL_BufferCacheHitRatio"); 
  #&SysPerfSample("\\${counter_prefix}:General Statistics\\User Connections",   "SQL_UserConnections"); 
  #&SysPerfSample("\\${counter_prefix}:General Statistics\\Processes Blocked",  "SQL_ProcessesBlocked"); 

  #---- We have now collected all of the data

  &msg("Resulting hash sysperf:"); 
  foreach $this_key (sort keys %sysperf)
  {
    &msg("  Key=$this_key Value=$sysperf{$this_key}");
  }
 
  if ($sysperf{"MemAvailMBytes"} > 0) 
  {
    $mem_avail = $sysperf{"MemAvailMBytes"} + 0;  
    &msg("Set mem_avail from sysperf hash - mem_avail=$mem_avail"); 
  } 
  
  $sp_elapsed = time - $sp_start; 
  &msg("SysPerf - Done - Elapsed: $sp_elapsed second(s)");
}

#------------------------------------------------------------------------------
sub SysPerfBulk
#------------------------------------------------------------------------------
{
  &msg("SysPerfBulk - Starting");
  $spb_stime = time; 

  #---- Each DB seems to have a different way of specifying the first part of the verb 
  #---- So by default we will use "SQLServer:", but we will query sys.dm_os_performance_counters to get the real value
  $counter_prefix = "SQLServer";   #---- Default
  undef %sysperf_bulk;
  undef %perf_counters;
  my $perf_probeout; 
  $debug_perfbulk = 0; 

  if ($vershort > 80)
  {
    &connectDB if ! $dbh; 
    &msg("Selecting from dm_os_performance_counters to obtain counter_prefix"); 
    $sql = "select top 1 OBJECT_NAME from sys.dm_os_performance_counters";
    $sth = $dbh->prepare("$sql") || &dbError("Prepare1954");
    $sth->execute                || &dbError("Execute_Select1955");
    while (@Data = $sth->fetchrow_array)
    {
      $counter_string = $Data[0];
    }
    $sth->finish || &dbErrorOK("Finish1960");
    $counter_prefix_raw = (split(":",$counter_string)) [0];
    $counter_prefix = $counter_prefix_raw if $counter_prefix_raw;
    &msg("Resulting counter_string=$counter_string counter_prefix_raw=$counter_prefix_raw"); 
  }
  &msg("Using counter_prefix=$counter_prefix"); 

  #---- Now collect the data from the windoze performance counters

  #---- NEW - Collect from parms in CFG file - Write all out to $sps_file
  $sps_file_in  = "dbamon_typeperf_in.txt"; 
  open(SPSFILE,">$sps_file_in") || &error("Cannot open output $sps_file_in - $!"); 

  &msg("Writing Perf Counters from CFG file to $sps_file_in"); 
  foreach (sort keys %cfg_perfcounter)
  {
    $counter     = $_; 
    $dbamon_verb = $cfg_perfcounter{$_};

    if ($counter =~ /SQLSERVERPREFIX/) { $counter =~ s/SQLSERVERPREFIX/$counter_prefix/; } 
    
    &msg("  Found: counter=$counter dbamon_verb=$dbamon_verb - Writing to $sps_file_in") if $debug_perfbulk;
    print SPSFILE "$counter\n"; 
    $counter =~ s/^\\//; 
    $perf_counters{$counter} = $dbamon_verb;
  }
  close(SPSFILE); 

  #---- Decide which command to use - typeperf or dbamon_mssql_tpr 
  $typeperf_cmd = "typeperf.exe"; 
  &msg("Verifying that typeperf exists"); 
  open(SPSCMD,"$typeperf_cmd 2>&1 |"); 
  while(<SPSCMD>)
  {
    if (/is not recognized as an internal or external command/i) 
    { 
      $typeperf_cmd = "dbamon_mssql_tpe.exe"; 
      &msg("  typeperf NG - Will use typeperf_cmd=$typeperf_cmd"); 
    } 
  }
  close(SPSCMD); 
  &msg("We will be using typeperf_cmd=$typeperf_cmd"); 

  #---- Run typeperf and collect the results
  $sps_first_line = 0; 
  $sps_cmd = "$typeperf_cmd -cf $sps_file_in -sc 1 -si 3 "; 

  &msg("Running 3 second sample of performance counters - sps_cmd=$sps_cmd"); 

  open(SPSCMD,"$sps_cmd 2>&1 |"); 
  while(<SPSCMD>)
  {
    next if /^\s$/; 
    next if /^Warning:/; 
    print;
    
    @sps_words = split (/\,/,$_) ; 
    if (! $sps_first_line && /CSV /) 
    {
      $icnt = 0; 
      foreach $this_word (@sps_words)
      {
        $this_word =~ s/\n//g; 
        $this_counter = $this_word; 
        $this_counter =~ s/\\\\$host_uc\\//; 
        $this_counter =~ s/\"//g; 
        $icnt++; 
        $typeperf_keys{$icnt} = $this_counter; 
        &msg("  DEBUG - COUNTER - icnt=$icnt this_word=$this_word this_counter=$this_counter") if $debug_perfbulk; 
      }
      $sps_first_line = 1; 
      next; 
    }
    else   #---- There should just be 2 non-null lines
    {
      $icnt = 0; 
      foreach $this_word (@sps_words)
      {
        next if $this_word =~ /CSV /; 

        $this_word =~ s/\n//g; 
        $this_word =~ s/\"//g; 
        $icnt++; 
        $this_key         = $typeperf_keys{$icnt}; 
        $this_dbamon_verb = $perf_counters{$this_key}; 
        &msg("  DEBUG - VAL - icnt=$icnt this_word=$this_word this_key=$this_key this_dbamon_verb=$this_dbamon_verb") if $debug_perfbulk; 
        next if ! $this_dbamon_verb; 

        #---- Processing for counters with many decimal places
        if ($this_dbamon_verb =~ /MULT1000$/)
        {
          $this_dbamon_verb =~ s/MULT1000//; 
          $this_word = $this_word * 1000; 
        }
     
        $perf_probeout .= "$P SYSPERF: $this_dbamon_verb $this_word\n"; 
        $sysperf_bulk{$this_dbamon_verb} = $this_word; 
        $sysperf_bulk{$this_dbamon_verb} = sprintf("%.2f", $sysperf_bulk{$this_dbamon_verb}); 
      }
      last; 
    }
  }
  close(SPSCMD);

  #---- We have now collected all of the data

  my $sp_count_nonzero = 0; 
  &msg("Resulting hash sysperf_bulk:"); 
  foreach $this_key (sort keys %sysperf_bulk)
  {
    &msg("  Key=$this_key Value=$sysperf_bulk{$this_key}");
    $sp_count_nonzero++ if $sysperf_bulk{$this_key} > 0;   
  }
  &msg("SysPerfBulk - Count of non-zero values: $sp_count_nonzero"); 
  &msg("PHONEHOME: SysPerfBulk - Count of non-zero values: $sp_count_nonzero (< 3)") if $sp_count_nonzero < 3 && $osname ne "Win2000"; 

  #---- Write out the counters that we collected
  if ($perf_probeout) 
  { 
    &msg("PERF Probeout:\n$perf_probeout"); 
  }
 
  if ($sysperf_bulk{"MemAvailMBytes"} > 0) 
  {
    $mem_avail = $sysperf_bulk{"MemAvailMBytes"} + 0;  
    &msg("Set mem_avail from sysperf_bulk hash - mem_avail=$mem_avail"); 
  } 
  
  $spb_elapsed = time - $spb_stime; 
  &msg("SysPerfBulk - Done - Elapsed: $spb_elapsed second(s)");
}

#------------------------------------------------------------------------------
sub SysPerfSpecial
#------------------------------------------------------------------------------
{
  &msg("SysPerfSpecial - Starting");

  #---- #1 - Blocked processes GROSS
  $sql = "select count(*) from master..sysprocesses (nolock) where blocked != 0"; 
  $sth = $dbh->prepare($sql) || &dbError("Prepare_L2556");
  $sth->execute || &dbErrorOK("Execute_L2557");
  while (@Data = $sth->fetchrow_array)
  {
    print "$P SYSPERF: SQL_BlockedProcsGross $Data[0]\n"; 
  }
  
  #---- #2 - Blocked processes BLOCKED BY SOMEONE ELSE
  $sql = "select count(*) from master..sysprocesses (nolock) where blocked != 0 and blocked != spid"; 
  $sth = $dbh->prepare($sql) || &dbError("Prepare_L2565");
  $sth->execute || &dbErrorOK("Execute_L2566");
  while (@Data = $sth->fetchrow_array)
  {
    print "$P SYSPERF: SQL_BlockedProcsOtherProc $Data[0]\n"; 
  }

  &msg("SysPerfSpecial - Done");
}

#------------------------------------------------------------------------------
sub SysPerfCompare
#------------------------------------------------------------------------------
{
  &msg("SysPerfCompare - Starting");
  my $zstring = ""; 
  my $zatleastone = 0; 

  &msg("Resulting hash sysperf:"); 
  foreach $this_key (sort keys %sysperf)
  {
    $zstring = sprintf("%-40s %12.2f", $this_key, $sysperf{$this_key});  
    &msg("  $zstring");
  }

  &msg("Resulting hash sysperf_bulk:"); 
  foreach $this_key (sort keys %sysperf_bulk)
  {
    $zstring = sprintf("%-40s %12.2f", $this_key, $sysperf_bulk{$this_key});  
    &msg("  $zstring");
    $zatleastone = 1 if $sysperf_bulk{$this_key} > 0;  
  }

  &msg("Timing:"); 
  &msg("  Elapsed BULK:  $spb_elapsed"); 
  &msg("  Elapsed Other: $sp_elapsed"); 

  &msg("Key Count:"); 
  my $count = keys %sysperf_bulk;
  my $countb = keys %sysperf;
  &msg("  KeyCount BULK:  $countb"); 
  &msg("  KeyCount Other: $count"); 

  if ($count == $countb && $zatleastone)
  {
    &msg("Everything looks GOOD - Equal key counts and at least 1 Bulk value - Copying sysperf_bulk to sysperf");   
    $sysperf = $sysperf_bulk; 
  }  
  elsif  ($count >= 3 && $countb >= 3)
  {
    &msg("Everything looks GOOD - Bulk and Other each have >= 3 keys - Copying sysperf_bulk to sysperf");   
    $sysperf = $sysperf_bulk; 
  }  
  elsif ($count == 0 && $countb == 0)
  {
    &msg("Everything looks GOOD - ?? both Bulk and Other each have 0 keys (I guess this is OK?)  - Copying sysperf_bulk to sysperf");   
    $sysperf = $sysperf_bulk; 
  }  
  else
  {
    &msg("PHONEHOME: Something is wrong - Sysperf element count != SysperfBULK"); 
  }

  &msg("SysPerfCompare - Done");
}

#------------------------------------------------------------------------------
sub SQLPerf
#------------------------------------------------------------------------------
{
  $perf_before_time = time; 
  $perf_interval = 5; 

  &msg("SQLPerf - Starting - perf_before_time=$perf_before_time");

  #---- Call perfSample BEFORE
  &perfSample("Before"); 
 
  #---- Sleep before taking 2nd snapshot
  &msg("  Sleeping perf_interval=$perf_interval"); 
  sleep $perf_interval; 

  #---- Call perfSample AFTER
  $perf_after_time = time; 
  &perfSample("After"); 

  #---- Calculations
  $perf_el_time = $perf_after_time - $perf_before_time;
  &msg("  SQLPerf - Calculating Deltas - perf_before_time=$perf_before_time perf_after_time=$perf_after_time perf_el_time=$perf_el_time");

  foreach $p_metric (sort keys %perf_after_all)
  {
    $p_before_val = $perf_before_all{$p_metric};
    $p_after_val  = $perf_after_all{$p_metric};
    $p_delta_val  = $p_after_val - $p_before_val; 
    $p_per_second = -1;  #---- In case of zero divide - Unlikely
    $p_per_second = sprintf("%.2f", ($p_delta_val / $perf_el_time)) if $perf_el_time > 0;

    if ($p_metric =~ /rate/i) 
    {
      &msg("  Metric=$p_metric p_before_val=$p_before_val p_after_val=$p_after_val p_delta_val=$p_delta_val p_per_second=$p_per_second");  
      print "$P PERF: $p_metric $p_per_second\n"; 
    }
    else
    {
      &msg("  Metric=$p_metric p_before_val=$p_before_val p_after_val=$p_after_val");  
      if ($p_before_val > $p_after_val) { $p_final_val = $p_before_val; } 
      else                              { $p_final_val = $p_after_val; } 
      print "$P PERF: $p_metric $p_final_val\n"; 
    }
  }

  #$perf_total_trans = $perf_after_transcount - $perf_before_transcount; 
  #$perf_total_transrate = sprintf("%.2f", ($perf_total_trans / $perf_el_time)) if $perf_el_time > 0;

  #&msg("SQLPerf - BEFORE - perf_before_time=$perf_before_time - perf_before_transcount=$perf_before_transcount");
  #&msg("SQLPerf - AFTER  - perf_el_time=$perf_el_time - perf_after_transcount=$perf_after_transcount");

  &msg("SQLPerf - Done");
}

# ----------------------------------------------------------------------
sub bkrun
# ----------------------------------------------------------------------
{
  $db     = $arg1;
  $method = $arg2;
  $cmd    = $arg3;
  print "$P bkrun: Running backup for DB=$db Method=$method\n";
  print "$P bkrun: Command before {DB}, {SQLPATH}, {PERLPATH} and backslash substitution: $cmd\n";

  # 24Jan04 - Now PERLPATH contains the path to the executable
  #  If they specified perl.exe in the CMD, that is wrong. 
  if ($cmd =~ /perl.exe/ && $cmd =~ /PERLPATH/)
  {
    &msg("NOTE: PERLPATH is wrong - It specifies perl.exe - modifying"); 
    $cmd =~ s/\/perl.exe//g;
    &msg("Modified Command after substitution: $cmd");
    $modified = 1;
  }

  $cmd =~ s/\{DB\}/\"$db\"/g;
  $cmd =~ s'\[BLANK\]' 'g;
  $cmd =~ s'\[HYPHEN\]'-'g;

  $cmd =~ s/\{SQLPATH\}/$SQLPATH/g;
  $cmd =~ s/\{PERLPATH\}/$PERLPATH/g;

  $cmd =~ s/\//\\/g;
  print "$P bkrun: Command after substitution: $cmd\n";

  #DEBUG START
  system("$PERLPATH C:\\dbatools\\socketmail.pl dbaadmin\@dba.it.thecompany.com dbaadmin\@dba.it.thecompany.com \"dbamon_mssql.pl Modified_CMD   PERLPATH=$PERLPATH host=$host modified=$modified cmd=$cmd [AUTO]\" $dbname/$backup_type") if $modified;

  system("$PERLPATH C:\\dbatools\\socketmail.pl dbaadmin\@dba.it.thecompany.com dbaadmin\@dba.it.thecompany.com \"dbamon_mssql.pl PERLPATH_Not_E PERLPATH=$PERLPATH host=$host modified=$modified cmd=$cmd [AUTO]\" $dbname/$backup_type") if ! -e $PERLPATH;
  #DEBUG END 

  if ($method eq "B") { $syscmd = "start/b \"MSSQL $db Backup\" $cmd"; }
  else                { $syscmd = "$cmd"; }

  &msg("$bkrun: About to issue: $syscmd");
  $s_time = time;
  $rc = system($syscmd);
  $e_time = time;
  $el_time = $e_time - $s_time;
  &msg("bkrun: Back from issuing: $syscmd - rc=$rc el_time=$el_time");

  $rc /= 256 if $rc != 0;
  if ($rc != 0)
  {
    print "$P | ERROR: Invoked Backup_Command: ended with non-zero rc=$rc\n";
  }
}

# ----------------------------------------------------------------------
sub bkhi
# ----------------------------------------------------------------------
{
  $dbname = $arg1; 

  $dbname =~ s/\[HYPHEN\]/\-/g;
  $dbname =~ s/\[BLANK\]/\ /g;
  
  # use strict;
  use DBI;

  # Attempt to connect to the master database on the server, to verify MSSQL availability.
  &connectDB;
  
  $dbh->do("use msdb")         || &dbError("Use_MSDB");

  if (!$dberror)
  {
    # Create a report of all backups for this DB
    if ($ver =~ "6.5")
    { 
      # Check backup age for this db
      $sql = "
       select
	  substring(database_name,1,16) as db,
	  backup_type,
	  backup_finish,
	  datediff(hour,backup_finish, getdate()) as hours_since
       from
	  sysbackuphistory 
       where
	  backup_type = 1 
	  and database_name = '$dbname'
	  and datediff(day, backup_finish, getdate()) <= 50
       order by
	  3 desc";
    }
    else
    {
      # Check backup age for this db
      $sql = "
	 select
	    substring(database_name,1,16) as db,
	    type,
	    backup_finish_date,
	    datediff(hour, backup_finish_date, getdate()) as hours_since
	 from
	    backupset
	 where
	    database_name = '$dbname'
	    --and datediff(day, backup_finish_date, getdate()) <= 50
	 order by
	    3 desc";
    }

    &msg("Running query of MSDB for Backup History Report for database $dbname");
    $sth = $dbh->prepare("$sql") 
      || &dbError("Prepare");
    $sth->execute
      || &dbError("Execute_Select");

    print "\n<table border=1 bgcolor=skyblue class=c1>
    <tr bgcolor=slate> 
      <th><font size=2>Backup<br>Type
      <th><font size=2>Timestamp
      <th><font size=2>Age<br>(Hours)
    \n";
    $count = 0;
    while (@Data = $sth->fetchrow_array)
    {
      ++$count;
      $db      = $Data[0];
      $bcktype = $Data[1];
      if ($bcktype eq "1") { $bcktype = "D" }; 
      $bckts   = $Data[2];
      $bckage  = $Data[3];

      if    ($bcktype eq "D") { $bt_out = "Database"; }
      elsif ($bcktype eq "I") { $bt_out = "Database Differential"; }
      elsif ($bcktype eq "L") { $bt_out = "Log"; }
      elsif ($bcktype eq "F") { $bt_out = "File"; }
      else                        { $bt_out = "?"; }

      $bt = "($bcktype)$bt_out";

      printf "<tr>
        <td><font size=2>%-4s </font>
        <td><font size=2>%-24s </font>
        <td align=right><font size=2>%-6s</font>
      \n",
      $bt,
      $bckts,
      $bckage
      ;
      $backup_count++;
    }
    print "</table>\n";
    print "\n<b>Total Backups: $backup_count</b>\n\n"; 
  }
}

# ----------------------------------------------------------------------
sub bkup
# ----------------------------------------------------------------------
{
  &msg("bkup - Starting"); 
  # use strict;
  use DBI;

  #---- Default backup method
  $backup_method = "UN";    #---- Default type of unknown

  if ($company eq "thecompany") 
  {
    if (-e "\\temp\\ALL.txt")
    {
      $all_log = "\\temp\\ALL.txt";
    }
    else
    {
      $all_log = "$SQLPATH\\dba_bkup_log\\ALL.txt";
    }

    &msg("Trying to determine backup method from $all_log"); 

    if (-e $all_log && ! -r $all_log)
    {
      system("$PERLPATH C:\\dbatools\\socketmail.pl dbaadmin\@dba.it.thecompany.com dbaadmin\@dba.it.thecompanthecompany.com \"dbamon_mssql.pl ALL Log Not Readable - all_log=$all_log [AUTO]\" $dbname/$backup_type") if $modified;
    }

    open(ALL_LOG, "$all_log") || print "$P | Cannot open $all_log - $!\n";
    $lines_read = 0;
    while (<ALL_LOG>)
    {
      $lines_read++;
      if (/Backup_Method:/)
      {
	@words  = split(' ',$_); 
	$backup_method = $words[$#words];
      }
    }
    close(ALL_LOG);
    &msg("Read $lines_read lines from $all_log - backup_method=$backup_method"); 
  }
  
  &msg("Attempting to determine backup type"); 
  #---- Check for CommVault (CV) backup method and SYMANTEC backup method
  #---    If file exists and has been updated within the last 4 days, then this DB is using CV
  $cv_log_file = "c:\\program files\\commvault\\simpana\\log files\\sqlida.log";
  $c2_log_file = "D:\\program files\\commvault\\simpana\\log files\\sqlida.log";
  $sy_log_dir  = "C:\\Program Files\\Symantec\\Backup Exec\\RAWS\\Data";
  $s2_log_file = "C:\\Program Files\\Symantec\\Backup Exec\\Logs\\dbrecover.log";
  &msg("Looking for these files:"); 
  &msg("  $cv_log_file"); 
  &msg("  $c2_log_file"); 
  &msg("  $sy_log_dir"); 
  &msg("  $s2_log_file"); 

  if (-e "$cv_log_file")
  {
    $cv_file_mtime    = ( stat $cv_log_file )[9];
    $cv_file_age_days = (time - $cv_file_mtime) / 86400;
    if ($cv_file_age_days <= 30) { $backup_method = "CV"; }
    &msg("Found File - cv_file_mtime=$cv_file_mtime cv_file_age_days=$cv_file_age_days backup_method=$backup_method");
  }
  elsif (-e "$c2_log_file")
  {
    $c2_file_mtime    = ( stat $c2_log_file )[9];
    $c2_file_age_days = (time - $c2_file_mtime) / 86400;
    if ($c2_file_age_days <= 30) { $backup_method = "CV"; }
    &msg("Found File - c2_file_mtime=$c2_file_mtime c2_file_age_days=$c2_file_age_days backup_method=$backup_method");
  }
  elsif (-e "$sy_log_dir")
  {
    $sy_dir_mtime    = ( stat $sy_log_dir )[9];
    $sy_dir_age_days = (time - $sy_dir_mtime) / 86400;
    if ($sy_dir_age_days <= 14) { $backup_method = "SY"; }
    &msg("Found File - sy_dir_mtime=$sy_dir_mtime sy_dir_age_days=$sy_dir_age_days backup_method=$backup_method");
  }
  elsif (-e "$s2_log_file")
  {
    $s2_file_mtime    = ( stat $s2_log_file )[9];
    $s2_file_age_days = (time - $s2_file_mtime) / 86400;
    if ($s2_file_age_days <= 14) { $backup_method = "SY"; }
    &msg("Found File - s2_file_mtime=$s2_file_mtime s2_file_age_days=$s2_file_age_days backup_method=$backup_method");
  }

  &msg("Results - backup_method=$backup_method");

  #---- Attempt to connect to the master database on the server, to verify MSSQL availability.
  &connectDB;

  if (!$dberror)
  {
    print "$P | Checking for database master\n";

    $sth = $dbh->do("use master") || &dbError("Use_Master");

    # Read from master
    $sql = "select name from sysdatabases order by 1; ";
    $sth = $dbh->prepare("$sql") || &dbError("Prepare");
    $sth->execute || &dbError("Execute_Select");
    while (@Data = $sth->fetchrow_array)
    {
      $dbname = $Data[0];
      push(@dblist,$dbname);
      ++$numdbs;
    }

    # Process DB names
    foreach $dbname (@dblist)
    {
      $dbname_out  = $dbname;
      $dbname_out  =~ s/\s/[BLANK]/g;
      $dbname_out  =~ s/\-/[HYPHEN]/g;

      $dbname_out2 = $dbname_out; 
      $dbname_out2 = substr($dbname_out2,0,40) . "..." if length($dbname_out2) > 40; 

      if ($ver =~ "6.5")
        { 
           # Check backup age for this db
           $sql = "
       	   select
	      substring(database_name,1,16) as db,
	      max(backup_finish) as maxdate,
	      datediff(hour, max(backup_finish), getdate()) as hours_since
	   from
              sysbackuphistory 
	   where
	      backup_type =1
	      and database_name = '$dbname'
	   group by
	      database_name
           order by database_name, 3 desc";
        }
      else
        {
           # Check backup age for this db
           $sql = "
       	   select
	      substring(database_name,1,16) as db,
	      max(backup_finish_date) as maxdate,
	      datediff(hour, max(backup_finish_date), getdate()) as hours_since
	   from
              backupset
	   where
	      type in ('D','I')
	      and database_name = '$dbname'
	   group by
	      database_name 
           order by 1, 3 desc";
        }

      $dbh->do("use msdb")         || &dbError("Use_MSDB2");
      $sth = $dbh->prepare("$sql") || &dbError("Prepare");
      $sth->execute                || &dbError("Execute_Select");
      $bckfound = 0;
      while (@Data = $sth->fetchrow_array)
      {
        $bckfound = 1;
	$db      = $Data[0];
	$bckts   = $Data[1];
	$bckage  = $Data[2];

        #---- Get rid of spaces
        ##$bckts =~ s/\s/_/g;

        print "BckAge: $dbname_out $bckage $bckts $backup_method \n";
      }
      if (!$bckfound)
      {
	$db      = $dbname;
        print "BckAge: $dbname_out -1\n";
      }

      #---- Create backup report
      $sql = "select
        substring(database_name,1,16) as db,
        type,
        max(backup_finish_date) as maxdate,
        datediff(hour, max(backup_finish_date), getdate()) as hours_since
      from
        backupset
      where
	database_name = '$dbname'
      group by
        database_name,
        type
      order by 
        1, 2";
      $dbh->do("use msdb")         || &dbError("Use_MSDB2");
      $sth = $dbh->prepare("$sql") || &dbError("Prepare");
      $sth->execute                || &dbError("Execute_Select");
      $bckfound = 0;

      print "BckDet: <tr bgcolor=cadetblue><td colspan=9>$fs &nbsp; <b> Database: $dbname_out </b>\n";

      while (@Data = $sth->fetchrow_array)
      {
	$db             = $Data[0];
	$bck_type       = $Data[1];
	$bck_maxts_raw  = $Data[2];
	$bck_agehr      = $Data[3];
    
        #---- 2007-11-15 11:46:23.000
        $bck_maxts = substr($bck_maxts_raw,0,10);
        $bck_maxts =~ s/\-/\//g; 
        $bck_maxts .= "-" . substr($bck_maxts_raw,11,8);
        $bck_agedy = sprintf("%.2f", ($bck_agehr / 24));

        if    ($bck_type eq "D") { $bck_type_out = "(D) Database Full"; }
        elsif ($bck_type eq "I") { $bck_type_out = "(I) Database Differential"; }
        elsif ($bck_type eq "L") 
        { 
          $bck_type_out = "(L) Transaction Log"; 
          print "BckAgeTLOG: $dbname_out $bck_agehr $bck_maxts_raw $backup_method\n"; 
        }
        else                     { $bck_type_out = $bck_type; }

        ##print "BckDetTEST: $dbname_out bck_type=($bck_type) bck_maxts=($bck_maxts) bck_agehr=($bck_agehr)\n";
        print "BckDet: <tr><td>$fs <b> $dbname_out2 </b> <td>$fs $bck_type_out<td>$fs $bck_maxts<td align=right>$fs $bck_agehr<td align=right>$fs $bck_agedy\n";
      }
    }
  }

  #---- If BKTYPE is still UN, then try to figure out what's goin' on! 
  if ($company eq "CI" && $backup_method eq "UN" && 1 == 0)     #  <<<<<< DUMMIED OUT
  {
    #---- If    BACKUP_DEVICE is C: format, then SQL native backups
    #---- ElsIf BACKUP_DEVICE is \\ format, then SQL native backups
    #---- ElsIf NAME is NULL and UC user_name like '%TSERV%' then SYMANTEC backup 
    $sql = "SELECT
      ba.backup_start_date,
      ba.database_name,
      ba.type,
      bm.physical_device_name AS backup_device,
      ba.NAME,
      ba.USER_NAME,
      ba.DESCRIPTION,
      ba.backup_size
    FROM [msdb].[dbo].[backupset] ba,
	 [msdb].[dbo].[backupmediafamily] bm
    where
      ba.media_set_id = bm.media_set_id
      and ba.backup_start_date > dateadd(dd,-10,getdate())
      AND ba.type = 'D'
    "; 
    $sth = $dbh->prepare("$sql") || &dbError("Prepare");
    $sth->execute                || &dbError("Execute_Select");

    $bt_type_sql      = 0;
    $bt_type_symantec = 0; 

    while (@Data = $sth->fetchrow_array)
    {
      $bt_timestamp   = shift @Data;
      $bt_dbname      = shift @Data;
      $bt_type        = shift @Data;
      $bt_bck_device  = shift @Data;
      $bt_name        = shift @Data;
      $bt_user        = shift @Data;
      $bt_desc        = shift @Data;
      $bt_size        = shift @Data;

      if (substr($bt_bck_device,1,1) eq ":")            { $bt_type_sql = 1; } 
      elsif ($bt_bck_device =~ // )                 { $bt_type_sql = 1; } 
      elsif ($bt_name eq "" && uc($bt_user) =~ /TSERV/)  { $bt_type_symantec = 1; } 
      else  { } 
    }

    if    ($bt_type_symantec) { $backup_method = "SY"; }
    elsif ($bt_type_sql)      { $backup_method = "SQL"; }
    else    { } 
    print "$P Backup_Method: $backup_method\n"; 
  }
}

# ----------------------------------------------------------------------
sub backupJobs
# ----------------------------------------------------------------------
{
  &msg("Querying backup jobs ");
 
  if (! defined $dbh)
  {
    ##$dbh = DBI->connect("DBI:ODBC:$driver",'guest','' ) || &dbError("Connect_To_DB");
    &connectDB;
  }

  if ($sth)
  {
    $sth->finish();
  }

  $sth = $dbh->do("use msdb") || &dbError("Use_Master");

  $sql = "select 
      j.name, 
      s.command,
      j.enabled
    from  
      sysjobs j,
      sysjobsteps s
    where 
      j.job_id = s.job_id 
     
    ;";

  $sth = $dbh->prepare("$sql") || &dbError("Prepare");
  $sth->execute                || &dbError("Execute_Select");
  while (@Data = $sth->fetchrow_array)
  {
    $jobname    = $Data[0];
    $jobcmd     = $Data[1];
    $jobenabled = $Data[2];

    ##&msg("DEBUG jobname=$jobname jobenabled=$jobenabled");

    $jobcmd  =~ s/'/QUOTE/g;
    
    $jobcmd  = substr($jobcmd,0,200) if length($jobcmd) > 200;
    $jobname = substr($jobname,0,75) if length($jobname) > 75;

    if    ($jobcmd  =~ /call_backup/)         { $jobtype = "BackupVeritas"; }
    elsif ($jobcmd  =~ /call_disk_backup.pl/) { $jobtype = "BackupDisk"; }
    elsif ($jobcmd  =~ /OBII_backup/)         { $jobtype = "BackupOB"; }
    elsif ($jobname =~ /^billtest/)           { $jobtype = "BBTEST"; }
    else                                      { $jobtype = "Other"; }

    print "$P BACKUPJOB: \$jobname='$jobname'\; \$jobcmd='$jobcmd'\; \$jobtype='$jobtype'\; \n";

    $rundisable = $arg1;  # Set by Run_Surgery: DBC parameter

    if ($rundisable)
    {
      if ($jobtype =~ /^Backup/ && $jobenabled)
      {
        push(@sqltorun,"exec sp_update_job \@job_name='$jobname',\@enabled=0");
      }
    }
  }

  $jobsupdated = 0;
  
  &msg("Looking for jobs to update");

  if ($sqltorun[0])
  {
    foreach $sql (@sqltorun)
    { 
      &msg("Running sp_update_job - sql=$sql");
      $dbh->do("use msdb")          || &dbError("Use_MSDB");
      ##$sth2->finish()             || &dbError("Finish");
      $sth2 = $dbh->prepare("$sql") || &dbError("Prepare");
      $sth2->execute                || &dbError("Execute_Disable");
    }
  }

  &msg("Done processing jobs");
}

# ----------------------------------------------------------------------
sub oldPerl
# ----------------------------------------------------------------------
{
  $pslist_loc = "C:\\dbatools\\pslist.exe";
  $oldperl_t  = 24;    #---- 1.27
  $oldperl_cnt = 0;
 
  if (! -e $pslist_loc)
  {
    &msg("pslist is not found at $pslist_loc - Skipping");
    return;
  }

  &msg("Looking for old Perl procesess that have been running for at least $oldperl_t hours under userid=$userid - pslist_loc=$pslist_loc");

  $cmd = "$pslist_loc > NUL 2>&1";
  &msg("Testing $pslist_loc to make sure that it works on this server - cmd=$cmd");
  if (system("$cmd") != 0)
  {
    &msg("pslist Error encountered - Skipping");
    return;
  }

  open(CMD, "$pslist_loc |") || &error("Cannot invoke $pslist_loc - $!");
  while (<CMD>)
  {
    @words  = split;
    $pname  = $words[0];
    $pid    = $words[1];
    $utime  = $words[6];
    $ktime  = $words[7];
    $etime  = $words[8];

    $hours = (split(":",$etime)) [0];

    next if $pname !~ /Perl/;
    next if $hours < $oldperl_t;

    # Get username
    open(CMD2, "pulist  |") || print "$P | Cannot invoke $pslist_loc - $!\n";;
    while (<CMD2>)
    {
      @words  = split;
      $pgm2   = $words[0];
      $pid2   = $words[1];
      $usr2   = $words[2];

      ##print "pgm2=$pgm2 pid2=$pid2 usr2=$usr2\n"; 
 
      next if $pid2 ne $pid;

      $user   = $words[2];
    }
    close(CMD2);

    next if index($user,$userid) == -1;

    print "$P OLDPERL: pid=$pid pname=$pname etime=$etime hours=$hours user=$user utime=$utime ktime=$ktime\n";
    $oldperl_cnt++;
  }
  close(CMD);
  &msg("Found $oldperl_cnt old perl processes"); 
}

# ----------------------------------------------------------------------
sub checkDoubleTake
# ----------------------------------------------------------------------
{
  &msg("checkDoubleTake - Starting - Checking DoubleTake (registry)"); 

  #---- See if it is installed
  %values="";
  $Register = "Software\\NSI Software\\Double-Take\\CurrentVersion";
  $HKEY_LOCAL_MACHINE->Open($Register,$hkey); 
  ##&msg("  hkey=$hkey"); 
  if ($hkey eq "")
  {
    &msg("  DoubleTake is NOT installed"); 
    return;
  }
  $hkey->GetValues(\%values) || &msg("  Cannot find DoubleTake in registry");
  foreach $value (keys(%values))
  {
    $RegType      = $values{$value}->[1];
    $RegValue     = $values{$value}->[2];
    $RegKey       = $values{$value}->[0];
    $doubletake_version = $RegValue if $RegKey eq "VersionInfo";
  }
  $hkey->Close();
  if ($doubletake_version eq "")
  {
    &msg("  DoubleTake is NOT installed"); 
    return;
  }
  &msg("  DoubleTake version $doubletake_version is installed"); 

  #---- Get MSSQL base dir
  %values="";
  $Register = "Software\\Microsoft\\MSSQLServer\\MSSQLServer\\Parameters";
  $HKEY_LOCAL_MACHINE->Open($Register,$hkey); 
  $hkey->GetValues(\%values);
  foreach $value (keys(%values))
  {
    $RegType       = $values{$value}->[1];
    $RegValue      = $values{$value}->[2];
    $RegKey        = $values{$value}->[0];
    $mssql_basedir = substr($RegValue,2) if $RegKey eq "SQLArg0";
  }
  $hkey->Close();

  $zraw = $mssql_basedir;
  $zlookfor = "\\"; 
  $zidx = rindex($zraw,$zlookfor); 
  $doubletake_dir = substr($zraw,0,$zidx);

  &msg("  mssql_basedir=$mssql_basedir doubletake_dir=$doubletake_dir"); 

  $doubletake_fileage_mins = 999999; 
  $doubletake_file = "$doubletake_dir\\DBAmon_Timestamp.txt"; 
  if (-e $doubletake_file) 
  {
    open(TS, "$doubletake_file") || print "$P | Cannot open input $doubletake_file - $!\n";;
    while (<TS>)
    {
      @words = split;
      $tsnum = $words[0];
    }
    $doubletake_fileage_mins = int((time - $tsnum) / 60); 
  }

  #---- See if doubletake is running
  $doubletake_status = "?";
  open(NET, 'net start|') || print "Error: Running net start";
  while(<NET>)
  {
    ##print;
    if (/Double-Take/)            { $doubletake_status="[Running]"; }
  }
  close(NET);

  print "DoubleTakeStatus:      $doubletake_status\n";
  print "DoubleTakeDir:         $doubletake_dir\n";
  print "DoubleTakeFile:        $doubletake_file\n";
  print "DoubleTakeFileAgeMins: $doubletake_fileage_mins\n";

  &msg("checkDoubleTake - Done"); 
}

# ----------------------------------------------------------------------
sub connectDBOld
# ----------------------------------------------------------------------
{
  &msg("connectDB() Connecting to ODBC driver=$driver as 'guest'"); 
  $dbh = DBI->connect("DBI:ODBC:$driver",'guest','' ) || &dbError("Connect_To_DB");

  #if ($host eq "wbbndba1")
  #{
  #  &msg("  Setting isolation level: SET TRANSACTION ISOLATION LEVEL READ UNCOMMITTED"); 
  #  $dbh->do("SET TRANSACTION ISOLATION LEVEL READ UNCOMMITTED") || &dbError("SET ISO LEVEL");
  #}
  &msg("  Setting isolation level: SET TRANSACTION ISOLATION LEVEL READ UNCOMMITTED"); 
  $dbh->do("SET TRANSACTION ISOLATION LEVEL READ UNCOMMITTED") || &dbError("SET ISO LEVEL");
}

# ----------------------------------------------------------------------
sub connectDB
# ----------------------------------------------------------------------
{
  &msg("connectDB() Connecting to ODBC without driver (dynamic) as localhost\\$in_instance uid/pwd=null"); 

  if ($default_instance) 
  {
    $DSN = "driver={SQL Server};Server=localhost;uid=;pwd=;"; 
  }
  else
  {
    #---- OLD: if ($vershort <= 80) 
    if ($vershort <= 80 || -e "dbamon_mssql_Parm_ConnectToHostname.txt")
    {
      $DSN = "driver={SQL Server};Server=$host\\$in_instance;uid=;pwd=;"; 
    }
    else
    {
      $DSN = "driver={SQL Server};Server=localhost\\$in_instance;uid=;pwd=;"; 
    }
  }

  if ($dbh ne "") 
  {
    &msg("Already connected to DB - dbh is not null"); 
    return;
  }

  &msg("  connecting DSN=$DSN") ;
  $dbh = DBI->connect("dbi:ODBC:$DSN", 
  { PrintError => 1,
    RaiseError => 0 } ) || &error("DBConnect");

  $iso_level = "SET TRANSACTION ISOLATION LEVEL READ UNCOMMITTED"; 
  &msg("  Setting isolation level: $iso_level"); 
  $dbh->do($iso_level) || &error("SET ISO LEVEL");
  &msg("connectDB() Done"); 

  $dbh->do("use tempdb") || &error("SET ISO LEVEL");
  ##$sql = " IF OBJECT_ID (N'tempdb.dbo.dateconvert_19', N'FN') IS NOT NULL DROP FUNCTION dbo.dateconvert_19; ";
  $sql = " IF OBJECT_ID (N'dbo.dateconvert_19', N'FN') IS NOT NULL DROP FUNCTION dbo.dateconvert_19; ";
  ##&msg("Running SQL=$sql");
  $sth = $dbh->do($sql) || &error("DROP_F1");

  $dbh->do("use tempdb") || &error("SET ISO LEVEL");
  #  -- 2005+ only?: WITH EXECUTE AS CALLER
  ##$sql = "CREATE FUNCTION dbo.dateconvert_19 (\@DATE datetime2)
  $sql = "CREATE FUNCTION dbo.dateconvert_19 (\@DATE datetime)
    RETURNS varchar(19)
    AS
    BEGIN
	 DECLARE \@RESULT varchar(19);
	 SET \@RESULT = CONVERT(varchar(10), \@DATE, 111) +
		  '-' + CONVERT(varchar(8),  \@DATE, 108);
	 RETURN(\@RESULT);
    END
  ";
  ##&msg("Running SQL=$sql");
  $sth = $dbh->do($sql) || &error("Create_F1");

  $connected_to_db = 1; 

  #---- Query SQL EDITION now, since it will be needed for other checks
  &msg("Querying SQL for patchlevel and edition"); 
  $dbh->do("use master")                || &dbError("Use_Master");
  $sth = $dbh->prepare("
    SELECT SERVERPROPERTY('productversion'), 
    SERVERPROPERTY ('productlevel'), 
    SERVERPROPERTY ('edition')
  ") || &dbError("Prepare_SQLVER_L3034");
  $sth->execute                         || &dbError("Execute_Select");
  $sql_version = $sql_patchlevel = $sql_edition = "?"; 
  while (@Data = $sth->fetchrow_array)
  {
    $sql_version      = $Data[0]; 
    $sql_patchlevel   = $Data[1]; 
    $sql_edition      = $Data[2]; 
  }
  $sth->finish() || &dbError("Could not execute FINISH");
  &msg("Results - sql_version=$sql_version sql_patchlevel=$sql_patchlevel sql_edition=$sql_edition"); 

  &msg("connectDB - Done"); 
}

# ----------------------------------------------------------------------
sub test
# ----------------------------------------------------------------------
{
  &msg("test()"); 

  &msg("Setting isolation level: SET TRANSACTION ISOLATION LEVEL READ UNCOMMITTED"); 
  $dbh->do("SET TRANSACTION ISOLATION LEVEL READ UNCOMMITTED") || &dbError("Use_Master_1");


  $dbname = "master";

	&msg("Running exec sp_spaceused for dbname='$dbname'"); 
	$sth = $dbh->prepare("exec sp_spaceused") || &dbError("Prepare");
	$dberror = 0;
	$sth->execute || &dbErrorOK("Connecting to db=$dbname");
	if ($dberror)
	{
	  ++$numdbsoffline;
	  print "DB:     $dbname_out CONNECTERROR \n";
	  next;
	}
	while (@Data = $sth->fetchrow_array)
	{
	  ##print "Data0=$Data[0]\n";   # Test
	  ##@words = split ' ', $Data[0];
	  ##$verb = $words[0];
	  ##if ($verb eq $dbname)
	  if (uc $Data[0] eq uc $dbname)
	  {
	    $db_spacealloc = $Data[1];
	    chop $db_spacealloc ; chop $db_spacealloc ;chop $db_spacealloc ;
	    $db_spacefree  = $Data[2];
	    chop $db_spacefree ; chop $db_spacefree ;chop $db_spacefree ;
	    $db_spaceused  = $db_spacealloc - $db_spacefree;
	  }
	}

  &msg("db_spaceused=$db_spaceused"); 
}

# ----------------------------------------------------------------------
sub DCFRead 
# ----------------------------------------------------------------------
{
  #---- DCF (DBAmon control file) path
  $dcf_path = "dbamon_mssql_dcf_${in_instance}.txt";
  $dcf_debug = 1; 

  &msg("DCFRead - Starting - Reading dcf_path=$dcf_path dcf_debug=$dcf_debug"); 

  if (-e "$dcf_path") 
  {
    #---- Only require the file if it passes a perl syntax check
    if (system("perl -wc $dcf_path 2>&1") != 0)
    {
      &msg("Require file $dcf_path Perl compile error - Not requiring"); 
      # system("echp ' ' | mailx -s 'DBAmon | $host/$os $dcf_path Compile Error' dbaadmin\@dba.it.mycompany.com") if $company eq "mycompany";
    }
    elsif (-z $dcf_path)
    {
      &msg("Require file $dcf_path Is zero bytes - Not requiring"); 
    }
    else
    {
      require "$dcf_path"; 
    }
  }
  else
  {
    &msg("Controlfile file $dcf_path missing - DCFFILEMISSING"); 
    # system("echo ' ' | mailx -s 'DBAmon | $host/$os - DCF $dcf_path Missing' dbaadmin\@dba.it.mycompany.com") if $company eq "mycompany";
  }

  #---- TEMP DEBUG: show variable values
  if ($dcf_debug) 
  {
    &msg("DEBUG - DCF Values read:"); 
    foreach $dcf_key (sort keys %dcf_data)
    {
      &msg("  dcf_data{$dcf_key}=$dcf_data{$dcf_key}"); 
    }
  }

  if ($dcf_data{"LastCycleFinishNum"} > 0)
  {
    $zmins_since = sprintf("%.2f", (time - $dcf_data{"LastCycleFinishNum"}) / 60);
    &msg("DCFRead - Done - Minutes since completion of last iteration: $zmins_since"); 
  }
}

#------------------------------------------------------------------------------
sub timeToText
#------------------------------------------------------------------------------
{
  my $zz_ts = $_[0];
  my ($sec,$min,$hour,$mday,$mon,$year) = localtime($zz_ts);
  $year += 1900;
  $mon += 1;

  # Pad with 0's
  my $yy = sprintf("%04d",$year);
  my $mm = sprintf("%02d",$mon);
  my $dd = sprintf("%02d",$mday);
  my $hh = sprintf("%02d",$hour);
  my $mi = sprintf("%02d",$min);
  my $ss = sprintf("%02d",$sec);

  my $date = "$yy/$mm/$dd-$hh:$mi:$ss";
  return $date;
}

# ----------------------------------------------------------------------
sub DCFWrite 
# ----------------------------------------------------------------------
{
  &msg("DCFWrite - Starting - Writing dcf_path=$dcf_path"); 
  $dcf_debug = 1; 

  #---- Cycle finish variable
  my $z = time; 
  $dcf_data{"LastCycleFinishNum"}  = $z; 
  $dcf_data{"LastCycleFinishChar"} = &timeToText($z); 

  #---- Write out values
  foreach $dcf_key (sort keys %dcf_data)
  {
    $dcf_out .= "\$dcf_data{$dcf_key} = '$dcf_data{$dcf_key}';\n";
    $dcf_count++; 
  }
  ##&msg("DEBUG - dcf_count=$dcf_count dcf_out=$dcf_out"); 

  open(DCF, ">$dcf_path") || &msg("Cannot open output $dcf_path - $!");
  print DCF $dcf_out; 
  close(DCF);

  system("type $dcf_path") if $dcf_debug;

  &msg("DCFWrite - Done - Wrote $dcf_count controlfile entries"); 
}

# ----------------------------------------------------------------------
sub init
# ----------------------------------------------------------------------
{
  &msg("init - Starting"); 

  $stime = time;

  $verb = $ARGV[0];
  $arg1 = $ARGV[1];
  $arg2 = $ARGV[2];
  $arg3 = $ARGV[3];
  $arg4 = $ARGV[4];

  $WWW = 1 if $arg1 eq "WWW";

  if (system("srvinfo -ns >NUL 2>NUL") == 0) { $srvinfo_cmd = "srvinfo -ns"; }
  elsif ($ENV{'NTRESKIT'}) { $srvinfo_cmd = "$ENV{'NTRESKIT'}\\srvinfo -ns"; }
  elsif (-e "dbamon_mssql_di.exe") { $srvinfo_cmd = "dbamon_mssql_di.exe -ns"; }
  else
  {
    &msg("Hmmmm... Could not find srvinfo in default path and NTRESKIT not in env");
  }

  $system_root = $ENV{"SYSTEMROOT"};
  $root_fs = substr($system_root,0,2);

  #TEMP#
  #$temp_dir = "$root_fs\\temp\\dbamon";
  #if (!(-d $temp_dir))
  #{
  #  mkdir ("$temp_dir",0777);
  #}

  $driver = "DBAmonMSSQL";

  #---- Read dbamon_mssql.cfg file
  &readCfg;
  
  &msg("  host=$host verb=$verb arg1=$arg1 arg2=$arg2 arg3=$arg3 arg4=$arg4 - driver=$driver");
  &msg("init - Done"); 
}

# ----------------------------------------------------------------------
sub readCfg
# ----------------------------------------------------------------------
{
  &msg("readCfg - Starting"); 

  $file_cfg_local = "dbamon_mssql_local.cfg"; 
  $file_cfg = "dbamon_mssql.cfg"; 

  &msg("Checking existence of LOCAL CFG file $file_cfg_local"); 
  if (-e "$file_cfg_local" && -r "$file_cfg_local")
  {
    $file_cfg = $file_cfg_local; 
    &msg("LOCAL CFG File DOES exist - It will be used ($file_cfg)"); 
  }
  else
  {
    &msg("LOCAL CFG File does NOT exist - Default file will be used ($file_cfg)"); 
  }

  &msg("Reading - file_cfg=$file_cfg"); 

  $cnt_cfg = 0;
  $cfg_debug = 0; 

  if (-e "$file_cfg" && -r "$file_cfg")
  {
    open(CFG, "$file_cfg") || &error("Error: Count not open input file_cfg=$file_cfg - $!");
    while(<CFG>)
    {
      ##print;
      next if /^#/ || /^$/;
      $cnt_cfg++; 
      @words = split;
      if ($words[0] eq "CHKALERT" && $words[1] eq "E")
      {
        push (@cfg_chkalert_e, $words[2]); 
      }
      elsif ($words[0] eq "PERFCOUNTER" )
      {
	$zp1 = (split(/\|/,$_))[1];
	$zp2 = (split(/\|/,$_))[2];

        $cfg_perfcounter{$zp1} = $zp2; 
      }
      else
      {
        &msg("ERROR: Verb=$words[0] Noun1=$words[1] Is undefined"); 
      }
    }
    close(CFG); 
  }
  else
  {
    &msg("Config file $file_cfg does not exist or is not readable"); 
  }

  if ($cfg_debug)
  {
    if (! $WWW)
    {
      &msg("Resulting \@cfg_chkalert_e"); 
      foreach (@cfg_chkalert_e)
      {
	&msg("  $_");
      }

      &msg("Resulting \%cfg_perfcounter"); 
      foreach (sort keys %cfg_perfcounter)
      {
        $zp1 = $_; 
        $zp2 = $cfg_perfcounter{$_};
        print "key=($zp1),val=($zp2)\n";
      }
    }
  }

  #---- Test
  &msg("From CFG file, these SQL Log errors will be excluded:"); 
  foreach (@cfg_chkalert_e)
  {
    &msg("  $_");
  }

  &msg("readCfg - Done - Non-comment lines read: $cnt_cfg"); 
}

# ----------------------------------------------------------------------
sub getRegInfo
# ----------------------------------------------------------------------
{
  &msg("getRegInfo - Starting"); 
  $debug_reg1 = 0;

  #---- Find $hostname
  %values="";
  $Register = "System\\CurrentControlset\\Services\\tcpip\\Parameters";
  $HKEY_LOCAL_MACHINE->Open($Register,$hkey)|| &msg("Cannot open registry 1 $Register - $!");
  $hkey->GetValues(\%values);
  foreach $value (keys(%values))
  {
    $RegType      = $values{$value}->[1];
    $RegValue     = $values{$value}->[2];
    $RegKey       = $values{$value}->[0];
    $host = $RegValue if $RegKey eq "Hostname";
  }
  $hkey->Close();
  &msg("DEBUG_REG1 - host=$host") if $debug_reg1;

  %values="";
  $Register = "System\\CurrentControlset\\Services\\tcpip\\Parameters";
  $HKEY_LOCAL_MACHINE->Open($Register,$hkey)|| &msg("Cannot open registry 2 $Register - $!");
  $hkey->GetValues(\%values);
  foreach $value (keys(%values))
  {
    $RegType      = $values{$value}->[1];
    $RegValue     = $values{$value}->[2];
    $RegKey       = $values{$value}->[0];
    $domain = $RegValue if $RegKey eq "Domain";
    $compwork=$domain;
  }
  $hkey->Close();

  # Save Company name
  @words = split(/\./,$compwork);
  $nexttolast = $#words - 1;
  $company = lc $words[$nexttolast];
  &msg("DEBUG_REG1 - company=$company") if $debug_reg1;

  # Get registry entries for Perl 
  %values="";
  $Register = "Software\\Perl";
  $HKEY_LOCAL_MACHINE->Open($Register,$hkey)|| &msg("Cannot open registry 3 $Register - $!");
  $hkey->GetValues(\%values);
  foreach $value (keys(%values))
  {
    $RegType      = $values{$value}->[1];
    $RegValue     = $values{$value}->[2];
    $RegKey       = $values{$value}->[0];
    $PERLPATH = $RegValue if $RegKey eq "";
  }
  $hkey->Close();
  &msg("DEBUG_REG1 - PERLPATH=$PERLPATH") if $debug_reg1;

  $path_length = length($PERLPATH);
  if (substr($PERLPATH,$path_length - 1,1) eq "\\")
  {
    chop($PERLPATH)
  }
  $PERLPATH = "$PERLPATH\\bin";
  if ($PERLPATH !~ /exe$/)
  {
    $PERLPATH .= "\\perl.exe";
  }

  #---- 
  #---- Find SQL values: SQLPATH, ver, vershort, mssql64bit
  #----
  $SQLPATH_2000 = ""; 
  $SQLPATH_2005 = ""; 

  #---- Find SQL values (old way - 2000)
  %values="";
  $Register = "Software\\Microsoft\\mssqlserver\\setup";
  $openkey = 1;
  $HKEY_LOCAL_MACHINE->Open($Register,$hkey) || ($openkey = 0);
  if ($openkey)
  {
    $hkey->GetValues(\%values);
    foreach $value (keys(%values))
    {
      $RegType      = $values{$value}->[1];
      $RegValue     = $values{$value}->[2];
      $RegKey       = $values{$value}->[0];
      ##print "$RegKey $RegValue \n";
      $SQLPATH = $RegValue if $RegKey eq "SQLPath";
    }
    $hkey->Close();
  }
  else 
  { 
    &msg("Could not find registry entries - Old style MSSQL entries (this is OK) - Register=$Register"); 
  }
  #---- Get MS SQL version OLD STYLE
  %values="";
  $Register = "Software\\Microsoft\\MSSQLServer\\MSSQLServer\\CurrentVersion";
  $HKEY_LOCAL_MACHINE->Open($Register,$hkey)|| &msg("Cannot open registry $Register - $!");
  if ($hkey)
  {
    $hkey->GetValues(\%values);
    foreach $value (keys(%values))
    {
      $RegType      = $values{$value}->[1];
      $RegValue     = $values{$value}->[2];
      $RegKey       = $values{$value}->[0];
      $VER_2000 = $RegValue if lc($RegKey) eq "csdversion";
      $VER_2000 = $RegValue if lc($RegKey) eq "currentversion" && $VER_2000 eq "";
    }
    $hkey->Close();
  }

  #---- 2011/09/29 BB: We have a SQL 2000 NAMED INSTANCE which does not use either of the above formats. Let's try this:
  if ($SQLPATH eq "")
  {
    &msg("Looking for special SQL2000 Named Instance path?"); 
    %values="";
    $Register = "Software\\Microsoft\\Microsoft SQL Server\\$in_instance\\Setup";
    $openkey = 1;
    $HKEY_LOCAL_MACHINE->Open($Register,$hkey) || ($openkey = 0);
    if ($openkey)
    {
      $hkey->GetValues(\%values);
      foreach $value (keys(%values))
      {
	$RegKey    = $values{$value}->[0];
	$RegType   = $values{$value}->[1];
	$RegValue  = $values{$value}->[2];
	$SQLPATH   = $RegValue if $RegKey eq "SQLPath";
      }
      $hkey->Close();
    }
    &msg("Resulting SQLPATH=($SQLPATH)"); 

    #---- Get MS SQL version OLD STYLE
    %values="";
    $Register = "Software\\Microsoft\\Microsoft SQL Server\\$in_instance\\MSSQLServer\\CurrentVersion";
    $HKEY_LOCAL_MACHINE->Open($Register,$hkey)|| &msg("Cannot open registry $Register - $!");
    if ($hkey)
    {
      $hkey->GetValues(\%values);
      foreach $value (keys(%values))
      {
	$RegType      = $values{$value}->[1];
	$RegValue     = $values{$value}->[2];
	$RegKey       = $values{$value}->[0];
	$VER_2000 = $RegValue if lc($RegKey) eq "currentversion" && $VER_2000 eq "";
	$VER_2000 = $RegValue if lc($RegKey) eq "csdversion";
      }
      $hkey->Close();
    }
  }

  #---- Save this for later
  $SQLPATH_2000 = $SQLPATH; 
  $SQLPATH = ""; 

  #---- 2.01 - It may be that a server has both old and new format registry entries. We need to ALSO check for a named instance with this NAME.
  #----          if we don't find that, it's OK -- just use what we found above. 

  #---- 
  #---- Try NEW method (SQL2005, 2008, higher)
  #---- 

  #---- Find SQL instances
  %values="";
  $Register = "Software\\Microsoft\\Microsoft SQL Server\\Instance Names\\SQL";
  $openkey = 1;
  $HKEY_LOCAL_MACHINE->Open($Register,$hkey) || ($openkey = 0);
  if ($openkey)
  {
    $hkey->GetValues(\%values);
    foreach $value (keys(%values))
    {
      $RegKey       = $values{$value}->[0];
      $RegType      = $values{$value}->[1];
      $RegValue     = $values{$value}->[2];
      ##print "key=$RegKey value=$RegValue \n";
      $sql_instances{$RegKey} = $RegValue if $RegKey;
    }
    $hkey->Close();
    &msg("MSSQL Instances Found:"); 
    foreach $this_instance (sort keys %sql_instances)
    {
      &msg("  SQL_Instance: $this_instance  SQL_Info: $sql_instances{$this_instance}");
    }

    #---- Now ensure that the $in_instance for this execution WAS found in the list of instances
    if (! $sql_instances{$in_instance} ) 
    { 
      &msg("The Instance that you specified ($in_instance) Does NOT EXIST on this server - It could be that this instance is OLD"); 
    }
    else
    {
      #---- If we are here, then this server DOES have a list of instances and this in_instance DOES exist
      $in_instance_key = $sql_instances{$in_instance}; 
      &msg("Using in_instance_key=$in_instance_key to find SQL version"); 

      #---- Get MS SQL version
      %values="";
      $Register = "Software\\Microsoft\\Microsoft SQL Server\\$in_instance_key\\MSSQLServer\\CurrentVersion";
      $HKEY_LOCAL_MACHINE->Open($Register,$hkey)|| die "Cannot open registry $Register - $!";
      $hkey->GetValues(\%values);
      foreach $value (keys(%values))
      {
	$RegType      = $values{$value}->[1];
	$RegValue     = $values{$value}->[2];
	$RegKey       = $values{$value}->[0];
	$VER_2005 = $RegValue if lc($RegKey) eq "csdversion";
	$VER_2005 = $RegValue if lc($RegKey) eq "currentversion" && $VER_2005 eq "";
      }
      $hkey->Close();

      #---- Get SQLPATH
      $SQLPATH = "";
      %values = "";
      $Register = "Software\\Microsoft\\Microsoft SQL Server\\$in_instance_key\\Setup";
      $HKEY_LOCAL_MACHINE->Open($Register,$hkey)|| die "Cannot open registry $Register - $!";
      $hkey->GetValues(\%values);
      foreach $value (keys(%values))
      {
	$RegType      = $values{$value}->[1];
	$RegValue     = $values{$value}->[2];
	$RegKey       = $values{$value}->[0];
	$SQLPATH = $RegValue if lc($RegKey) eq "sqlpath";
      }
      $hkey->Close();
    }
  }
  else 
  { 
    &msg("Could not read registry to get List of MSSQL instances - This could be an OLD server - Register=$Register"); 
  }
  #---- Save this for later
  $SQLPATH_2005 = $SQLPATH; 

  #---- Decide which value to use for SQLPATH
  &msg("After REG samples - SQLPATH_2000=$SQLPATH_2000 SQLPATH_2005=$SQLPATH_2005 - Deciding which one to use..."); 
  if    ($SQLPATH_2005) 
  { 
    $SQLPATH = $SQLPATH_2005; 
    $ver     = $VER_2005; 
    &msg("Using SQLPATH_2005 SQLPATH value"); 
  }
  elsif ($SQLPATH_2000) 
  { 
    $SQLPATH = $SQLPATH_2000; 
    $ver     = $VER_2000; 
    &msg("Using SQLPATH_2000 SQLPATH value"); 
  }
  else  { &error("Could NOT FIND SQLPATH from registry"); }

  #---- See if we can see if this is 64bit mode
  #---- I will try to make this work from the MASTER DB (that would be better)
  #$key = "HKEY_LOCAL_MACHINE\\software\\wow6432node\\microsoft\\microsoft sql server\\100\\tools\\setup";
  #foreach (`reg query \"$key\" `)
  #{
  #  #print;
  #  @words = split;
  #  if (/64-bit/) { $mssql64bit = 1; }
  #}
  #&msg( "DEBUG mssql64bit=$mssql64bit"); 
     
  if (length((split(/\./,$ver)) [2]) == 3)
  {
    $ver = (split(/\./,$ver)) [0] . '.' .
	   (split(/\./,$ver)) [1] . '.' .
	   "0" . (split(/\./,$ver)) [2];
    &msg("DEBUG = Padding right node of version with 0 - ver=$ver"); 
  }

  #---- If after this we don't have SQLPATH, look for the values in this file
  $sqlpathfile = "c:\\SQLPATH.txt";
  if ($SQLPATH eq "" && -e "$sqlpathfile")
  {
    &msg("SQLPATH Being set from file $sqlpathfile"); 
    chop($SQLPATH = `type $sqlpathfile`); 
    &msg("New SQLPATH=$SQLPATH"); 
  }

  $vershort = (split(/\./,$ver)) [0] . '.' .
	      (split(/\./,$ver)) [1]; 
  $vershort =~ s/\.//;

  &msg("DEBUG - vershort before: $vershort"); 
  #---- VERSHORT could be in the format 800, 900 or 1000, etc.. edit that to be 80, 90, 100, 105
  if    ($vershort == 800)  { $vershort = 80; }
  elsif ($vershort == 900)  { $vershort = 90; }
  elsif ($vershort == 1000) { $vershort = 100; }
  elsif ($vershort == 1050) { $vershort = 105; }
  elsif ($vershort == 1100) { $vershort = 110; }
  &msg("DEBUG - vershort after: $vershort"); 

  #---- Format VERSION
  #if    ($ver =~ /^8.0/)      { $ver_major = "2000"; }
  #elsif ($ver =~ /^9.0/)      { $ver_major = "2005"; }
  #elsif ($ver =~ /^10.0/)     { $ver_major = "2008"; }
  #elsif ($ver =~ /^10.5/)     { $ver_major = "2008R2"; }
  #else                        { $ver_major = ""; } 
  ##
  #if    ($ver eq "8.00.0384") { $ver_sp = "SP1"; }
  #elsif ($ver eq "8.00.0534") { $ver_sp = "SP2";  }
  #elsif ($ver =~ /9.00.2047/) { $ver_sp = "SP1";  }
  #elsif ($ver =~ /9.00.2047/) { $ver_sp = "SP1";  }
  #elsif ($ver =~ /10.0.2531/) { $ver_sp = "SP1";  }
  #else                        { $ver_sp = ""; } 
  ##
  #$ver = "$ver($ver_major($ver_sp))"; 

  &error("Could not find SQLPATH") if ! $SQLPATH;

  $SQLPATH =~ s/\x0//g if $SQLPATH;    #---- Remove any NULLS !!!
  &msg("SQLPATH=($SQLPATH) PERLPATH=$PERLPATH ver=$ver WWW=$WWW userid=$userid vershort=$vershort mssql64bit=$mssql64bit");
  &msg("getRegInfo - Done - company=$company"); 
}

# ----------------------------------------------------------------------
sub driveCheck
# ----------------------------------------------------------------------
{
  &msg("driveCheckWMI - Loading DRIVE data into hash"); 

  use Win32::OLE('in'); 
  use constant wbemFlagReturnImmediately => 0x10; 
  use constant wbemFlagForwardOnly => 0x20; 

  $computer = ".";
  $objWMIService = Win32::OLE->GetObject
      ("winmgmts:\\\\$computer\\root\\CIMV2");  #--- or die "WMI connection failed.\n";
 
  if (! $objWMIService)
  {
    &msg("objWMIService is NULL - Error=($!) - Running driveCheckCMD instead");
    &driveCheckCMD;
    return;
  }

  $colItems = $objWMIService->ExecQuery 
      ("SELECT deviceid, size, freespace, status, statusinfo FROM Win32_LogicalDisk where drivetype = 3","WQL",wbemFlagReturnImmediately | wbemFlagForwardOnly); 
   
  $drive_count = 0;
  $tot_mb_size = 0;
  $tot_mb_free = 0;
  undef %disks;

  foreach my $objItem (in $colItems) 
  { 
    #print "Device ID: $objItem->{DeviceID}\n"; 
    #print "Free Space: $objItem->{FreeSpace}\n"; 
    #print "Size: $objItem->{Size}\n"; 

    $drive = $objItem->{DeviceID}; 
    $drive =~ s/\://;
    $disks{"$drive"} = 1;

    $disk_size_mb{"$drive"} = sprintf("%.2f", ($objItem->{Size}      / (1024*1024))  ); 
    $disk_free_mb{"$drive"} = sprintf("%.2f", ($objItem->{FreeSpace} / (1024*1024))  ); 
    $disk_used_mb{"$drive"} = sprintf("%.2f", ($disk_size_mb{"$drive"} - $disk_free_mb{"$drive"}));

    $disk_size_gb{"$drive"} = sprintf("%.2f", ($objItem->{Size}      / (1024*1024*1024))  ); 
    $disk_free_gb{"$drive"} = sprintf("%.2f", ($objItem->{FreeSpace} / (1024*1024*1024))  ); 
    $disk_used_gb{"$drive"} = sprintf("%.2f", ($disk_size_gb{"$drive"} - $disk_free_gb{"$drive"}));

    $disk_pct_full{"$drive"} = sprintf("%.2f", 100 * ($disk_used_gb{"$drive"} / $disk_size_gb{"$drive"})) if  $disk_size_gb{"$drive"} > 0;
    $drive_count++; 
    $tot_mb_size += $disk_size_mb{"$drive"}; 
    $tot_mb_free += $disk_free_mb{"$drive"}; 
  } 

  &msg("  Resulting Disk info:");
  foreach $zdrive (sort keys %disks)
  {
    &msg("    Disk: $zdrive - Size=$disk_size_gb{$zdrive} Free=$disk_free_gb{$zdrive} Used=$disk_used_gb{$zdrive}  PctF=$disk_pct_full{$zdrive}"); 
  }

  &msg("driveCheckWMI - Done - tot_mb_size=$tot_mb_size tot_mb_free=$tot_mb_free"); 
}

# ----------------------------------------------------------------------
sub countCPUs
# ----------------------------------------------------------------------
{
  &msg("countCPUs - Starting - srvinfo_cmd=$srvinfo_cmd"); 

  if ($ENV{"NUMBER_OF_PROCESSORS"})
  {
    &msg("  Setting numcpus from ENV"); 
    $numcpus = $ENV{"NUMBER_OF_PROCESSORS"};
  }
  else
  {
    &msg("  Setting numcpus from $srvinfo_cmd"); 
    open(SRVINFO, "$srvinfo_cmd 2>&1 |") || &error("Error: Running $srvinfo_cmd");
    while(<SRVINFO>)
    {
      ##print;
      next unless /^CPU/;
      $numcpus++; 
    }
    &msg("  Resulting numcpus=$numcpus"); 
  }

  print "PHONEHOME: NUMCPUs is NULL" if ! $numcpus;
  &msg("countCPUs - Done - numcpus=$numcpus"); 
}

# ----------------------------------------------------------------------
sub driveCheckCMD
# ----------------------------------------------------------------------
{
  &msg("driveCheckCMD - Loading DRIVE data into hash - running srvinfo_cmd=$srvinfo_cmd"); 

  open(SRVINFO, "$srvinfo_cmd 2>&1 |") || &error("Error: Running $srvinfo_cmd");
  $drive_count = 0;
  $tot_mb_size = 0;
  $tot_mb_free = 0;
  undef %disks;
  while(<SRVINFO>)
  {
    ##print;
    @words = split;

    if (length($words[0]) == 2 && ($words[1] eq "FAT" || $words[1] eq "NTFS"))
    {
      $drive = substr($words[0],0,1);
      $disks{"$drive"} = $_;
      $disk_size_gb{"$drive"} = sprintf("%.2f",$words[2] / 1024);
      $disk_free_gb{"$drive"} = sprintf("%.2f",$words[3] / 1024);
      $disk_used_gb{"$drive"} = sprintf("%.2f", ($disk_size_gb{"$drive"} - $disk_free_gb{"$drive"}));
      $disk_size_mb{"$drive"} = sprintf("%.2f",$words[2] );
      $disk_free_mb{"$drive"} = sprintf("%.2f",$words[3] );
      $disk_used_mb{"$drive"} = sprintf("%.2f", ($disk_size_mb{"$drive"} - $disk_free_mb{"$drive"}));
      $disk_pct_full{"$drive"} = sprintf("%.2f", 100 * ($disk_used_gb{"$drive"} / $disk_size_gb{"$drive"})) if  $disk_size_gb{"$drive"} > 0;
      $drive_count++; 
      $tot_mb_size += $disk_size_mb{"$drive"}; 
      $tot_mb_free += $disk_free_mb{"$drive"}; 
    }
  }

  &msg("  Resulting Disk info:");
  foreach $zdrive (sort keys %disks)
  {
    &msg("    Disk: $zdrive - Size=$disk_size_gb{$zdrive} Free=$disk_free_gb{$zdrive} Used=$disk_used_gb{$zdrive}  PctF=$disk_pct_full{$zdrive}"); 
  }

  &msg("driveCheckCMD - Done - tot_mb_size=$tot_mb_size tot_mb_free=$tot_mb_free"); 
}

# ----------------------------------------------------------------------
sub checkFilegroups
# ----------------------------------------------------------------------
{
  &msg("checkFilegroups - Starting - Checking filegroups");

  $debug_fg = 0;

  if ($WWW)
  {
    $debug_fg = 0;
  }

  &driveCheck;
  &connectDB if ! $dbh; 

  #---- Get list of ALL DBs
  $dbh->do("use master") || &dbError("Use_Master");
  $sql = "sp_helpdb";
  $sth3 = $dbh->prepare($sql) || &error("Could not prepare sql=$sql");
  $sth3->execute() || &error("Could not execute sql=$sql");
  &msg("Listing DBs...") if $debug_fg; 
  while (@Data = $sth3->fetchrow_array)
  {
    &msg("  Found $Data[0]") if $debug_fg; 
    push(@db_list, $Data[0]);
  }

  #---- Process once for each DB

  &msg("Now processing each DB...") if $debug_fg; 
  foreach $this_db (@db_list)
  {
    $this_db_out = $this_db;
    $this_db_out =~ s/\s/[BLANK]/g;
    $this_db_out =~ s/\-/[HYPHEN]/g;
    &msg("  Processing this_db=$this_db this_db_out=$this_db_out") if $debug_fg; 

    $sql = "sp_helpfilegroup";
    $dbh->do("use \[$this_db\]") || &dbError("Use_${this_db}");
    $sth3 = $dbh->prepare($sql) || &error("Could not prepare sql=$sql");
    $sth3->execute() || &error("Could not execute sql=$sql");

    undef @fg_list;
    $selectcount = 0;
    ##&msg("  Listing filegroups..."); 
    while (@Data = $sth3->fetchrow_array)
    {
      $selectcount++;

      $fg_name    = $Data[0];
      $fg_groupid = $Data[1];
      $fg_filecnt = $Data[2];

      &msg("    Found fg_name=$fg_name fg_groupid=$fg_groupid fg_filecnt=$fg_filecnt") if $debug_fg;
      push(@fg_list, $fg_name);
    }
    $sth3->finish() || &dbError("Could not execute FINISH");
    &msg("    Found $selectcount filegroup(s)") if $debug_fg; 

    &msg("    Processing list of FGs") if $debug_fg; 
    foreach $this_fg (@fg_list)
    {
      &msg("      Processing this_fg=$this_fg") if $debug_fg; 
 
      $tot_pot_mb = $tot_size_mb = $tot_maxsize_mb = 0; 
      $tot_drive_list = "";
      $unl_flag = 0;
      $nogrow_flag = 0;

      $sql = "exec sp_helpfilegroup \@filegroupname='$this_fg'";
      $sth = $dbh->prepare("$sql") || &error("Prepare");
      $sth->execute || &error("Execute_Select");
      while(@Data = $sth->fetchrow_array or $sth->{odbc_more_results})
      {
        next if $#Data != 5;
	$fg_file_in_group = $Data[0];
	$fg_fileid        = $Data[1];
	$fg_filename      = $Data[2];
	$fg_size          = $Data[3];
	$fg_maxsize       = $Data[4];
	$fg_growth        = $Data[5];

        #---- In case we don't set this later
        $fg_maxsize_mb = -2;

        if ($fg_maxsize =~ /unlimited/i && $fg_growth eq "0 KB") 
        { 
          &msg("*** DB=$this_db_out FG=$this_fg - This FG is FIXED SIZE ***");   #---- if $debug_fg; 
          $nogrow_flag = 1;

          #$sth->finish || &dbError("Execute_Finish");
          #last;
        }
        elsif ($fg_maxsize =~ /unlimited/i) 
        { 
          &msg("*** DB=$this_db_out FG=$this_fg - This FG has at least 1 FILE with maxsize UNLIMITED ***") if $debug_fg; 
          $unl_flag = 1;

          #$sth->finish || &dbError("Execute_Finish");
          #last;
        }

        if ($fg_size =~ / KB$/) { $fg_size_mb = $fg_size; $fg_size_mb =~ s/\sKB$//; $fg_size_mb = sprintf("%.2f",($fg_size_mb / 1024)); }
        else { &error("Unknown UNITS 2584"); }

        #---- if ($unl_flag == 0)
        if ($unl_flag == 0 && $nogrow_flag == 0)
        {
	  if ($fg_maxsize =~ / KB$/) { $fg_maxsize_mb = $fg_maxsize; $fg_maxsize_mb =~ s/\sKB$//; $fg_maxsize_mb = sprintf("%.2f",($fg_maxsize_mb / 1024)); }
	  else { &error("Unknown UNITS 2588 - fg_maxsize=$fg_maxsize"); }
        }

        if ($unl_flag) { $fg_potential_growth_mb = -1; }  
        else           { $fg_potential_growth_mb = $fg_maxsize_mb - $fg_size_mb; }

        $fg_drive = uc substr($fg_filename,0,1); 

	&msg("        FG - Found fg_filename=$fg_filename") if $debug_fg; 
	&msg("        FG -   fg_size_mb=$fg_size_mb fg_maxsize_mb=$fg_maxsize_mb fg_potential_growth_mb=$fg_potential_growth_mb fg_drive=$fg_drive") if $debug_fg; 

        $tot_drive_list .= $fg_drive; 
        $tot_pot_mb     += $fg_potential_growth_mb;
        $tot_size_mb    += $fg_size_mb;
        $tot_maxsize_mb += $fg_maxsize_mb unless $unl_flag;

        $drive_pot_growth{$fg_drive} += $fg_potential_growth_mb if $fg_potential_growth_mb > 0; 
      }
      $sth->finish || &dbError("Execute_Finish");

      $tot_pct_full_vs_max = -1; 
      $tot_pct_full_vs_max = sprintf("%.2f", (($tot_size_mb / $tot_maxsize_mb) * 100)) if $tot_maxsize_mb > 0; 
      &msg("      END OF FG Processing - this_fg=$this_fg - tot_size_mb=$tot_size_mb tot_maxsize_mb=$tot_maxsize_mb tot_pct_full_vs_max=$tot_pct_full_vs_max") if $debug_fg; 
      push(@fg_display_list, sprintf("%-40s %-16s %16s %16s %14.2f %17s %8s",$this_db_out,$this_fg,$tot_size_mb,$tot_maxsize_mb,$tot_pct_full_vs_max,$tot_pot_mb, $tot_drive_list)); 
    }
  }

  if ($WWW)
  {
    foreach (@fg_display_list)
    {
      @words = split;
      $fg_db          = $words[0]; 
      $fg_fg          = $words[1]; 
      $fg_totsize     = $words[2]; 
      $fg_maxsize     = $words[3]; 
      $fg_pct_vs_max  = $words[4]; 
      $fg_pot_growth  = $words[5]; 
      $fg_drive       = $words[6]; 

      $fg_filecount = length($fg_drive); 

      if ($fg_pct_vs_max == -1) 
      { 
        $fg_pct_vs_max = "(At least 1 DF is MAXSIZE=Unlimited)"; 
        $color = "white"; 
        $fg_pot_growth  = "N/A"; 
      }
      else 
      { 
        if    ($fg_pct_vs_max < 80) { $color = "lime"; }
        elsif ($fg_pct_vs_max < 90) { $color = "yellow"; }
        else  { $color = "pink"; }
        
        $fg_pct_vs_max .= "%"; 
      } 

      if (! $fg_heading)
      {
	print "<br>\n";
	&msg("Filegroup Status (UC):");
	print "<br><table border=1 bgcolor=skyblue class=c1>\n";
	print "<tr bgcolor=slate>
	  <th><h5>Database 
	  <th><h5>Filegroup 
	  <th><h5>File Count 
	  <th><h5>Total<br> DB File<br> Size <br>(mB)
	  <th><h5>Max.<br> DB File<br> Size <br>(mB)
	  <th><h5>Percent Used<br>vs.<br>Max. Size </tr>
	  <th><h5>Potential Growth<br>This FG<br>(mB)
	  <th><h5>Drive List<br>For This<br>FG 
	\n";
        $fg_heading = 1;
      }

      print "<tr bgcolor=$color>
	    <td><h5> <b> $fg_db </b>
	    <td><h5> <b> $fg_fg </b>
	    <td align=right><h5> $fg_filecount
	    <td align=right><h5> $fg_totsize
	    <td align=right><h5> $fg_maxsize 
	    <td align=right><h5> $fg_pct_vs_max
	    <td align=right><h5> $fg_pot_growth
	    <td ><h5> $fg_drive
	  \n";
    }
    print "</table><br>\n"; 

    #---- Display drive growth potential
    foreach $this_drive (sort keys %drive_pot_growth)
    {
      if (! $fgdr_heading)
      {
	print "<br>\n";
	&msg("Potential Growth By Drive:");
	print "<br><table border=1 bgcolor=skyblue class=c1>\n";
	print "<tr bgcolor=slate>
	  <th><h5>Drive 
	  <th><h5>Potential Growth<br>This Drive<br>(mB)
	  <th><h5>Free Space <br>This Drive<br>(mB)
	  <th><h5>Status
	\n";
        $fgdr_heading = 1;
      }

      if ($drive_pot_growth{$this_drive} >  $disk_free_mb{$this_drive}) 
      {
        $color = "pink"; 
        $zmsg = "DB Files could grow beyond Free Space for drive $this_drive:"; 
      }
      else
      {
        $color = "lime"; 
        $zmsg = "DB Files could NOT grow beyond Free Space for drive $this_drive:"; 
      }
      print "<tr bgcolor=$color>
	    <td><h5> <b> $this_drive
	    <td align=right><h5> $drive_pot_growth{$this_drive} 
	    <td align=right><h5> $disk_free_mb{$this_drive} 
	    <td><h5> <b> $zmsg
	  \n";
    }
    print "</table><br>\n"; 
  }
  else
  {
    &msg("FG Final Results:"); 
    print "#Legend:                 DBName                                   FGName                TotalSizeMB        MaxSizeMB   PctFullvsMax PotentialGrowthMB    Drive\n"; 
    print "#Legend: (MaxSizeMB is -1 if any DFs are MAXUNLIMITED) PctFullvsMax\n"; 
    foreach $this_line (@fg_display_list)
    {
      print "$P FG_ITEM: $this_line\n"; 
    }
    &msg("FG Drive Growth Results:"); 
    print "#Legend:                 Drive     PotGrowthMB     FreeMB\n"; 
    foreach $this_drive (sort keys %drive_pot_growth)
    {
      print "$P FG_DRIVE_GROWTH: $this_drive $drive_pot_growth{$this_drive} $disk_free_mb{$this_drive}\n"; 
    }
  }

  &msg("checkFilegroups - Done");   
}

# ----------------------------------------------------------------------
sub checkDBMail
# ----------------------------------------------------------------------
{
  &msg("checkDBMail - Checking DBMail Status (SQL2005+ only)");

  if ($vershort < 90)
  {
    &msg("This check only works for DBMAIL (SQL >= 2005) - Returning"); 
    return;
  }

  #---- Check DBMAIL process status
  #&msg("  Checking to see if DBMAIL is ENABLED");
  ##$dbh->do("exec sp_configure 'show advanced options',1")       || &dbError("Do_4350");
  ##$sth = $dbh->prepare("exec sp_configure 'Database Mail XPs'") || &dbError("Prepare_L4350");
  #$sth = $dbh->prepare("exec sp_configure 'show advanced options',1; exec sp_configure 'Database Mail XPs'") || &dbError("Prepare_L4350");
  #$sth->execute || &dbErrorOK("Execute_4351");
  #while($href = $sth->fetchrow_hashref)
  #{
  #  &msg("  Found row"); 
  #  $dbmail_enabled = $href->{run_value}; 
  #}
  #&msg("  Resulting dbmail_enabled=$dbmail_enabled");

  #---- Look at last line of DBMAIL log table
  $sql = "SELECT
      sent_date,
      sent_status
    FROM
      msdb..sysmail_allitems (NOLOCK)
    WHERE
      sent_date = (SELECT MAX(sent_date) FROM msdb..sysmail_allitems)
    ORDER BY
      1
  "; 
  $sth = $dbh->prepare($sql) || &dbError("Prepare_L3546");
  $sth->execute || &dbErrorOK("execute helpdb db=$dbname");
  $mail_row_cnt = 0; 
  $sent_status = ""; 
  while($href = $sth->fetchrow_hashref)
  {
    $mail_row_cnt++;
    $sent_date    = $href->{sent_date}; 
    $sent_status  = $href->{sent_status}; 
    &msg("  Found sent_date=$sent_date sent_status=$sent_status");
  }
  $sent_status = "?" if $sent_status eq ""; 

  #---- Check DBMAIL process status
  if ($sent_status ne "?")
  {
    &msg("  Checking DBMAIL status by running sysmail_help_status_sp");
    $sth = $dbh->prepare("EXEC msdb.dbo.sysmail_help_status_sp;") || &dbError("Prepare_L4382");
    $sth->execute || &dbErrorOK("Execute_4383");
    $dbmail_status = "?"; 
    while($href = $sth->fetchrow_hashref)
    {
      $dbmail_status  = $href->{Status}; 
    }
    &msg("  Resulting dbmail_status=$dbmail_status");
  }

  if    ($sent_status eq "?")         { $zstat = "OK - Possibly_Not_Configured"; }
  elsif ($sent_status eq "sent")      { $zstat = "OK - (sent_status=$sent_status)"; }
  elsif ($dbmail_status eq "STOPPED") { $zstat = "NG - DBMAIL_Is_Stopped_(sent_status=$sent_status)"; }
  else                                { $zstat = "NG - (sent_status=$sent_status)"; }

  &msg("  Most recent status: $zstat"); 
  print "$P DBMAIL_STATUS: $zstat\n"; 

  &msg("checkDBMail - Done - Found $mail_row_cnt rows in SYSMAIL_ALLITEMS");
}

# ----------------------------------------------------------------------
sub checkLRBackup
# ----------------------------------------------------------------------
{
  &msg("checkLRBackup - Looking for LONG RUNNING BACKUPs");

  $sql = "SELECT
    DATEDIFF(hh,login_time,GETDATE()) AS Process_Age_Hours
    ,spid as SPID
  --  ,login_time
  --  ,last_batch
  --  ,open_tran
  --  ,status
  --  ,blocked
  --  ,lastwaittype
  --  ,cmd
  --  ,cpu
  --  ,physical_io
  FROM master..sysprocesses
  WHERE
    UPPER(cmd) LIKE '%BACKUP%'
    --AND DATEDIFF(hh,login_time,GETDATE()) >= 10
  ORDER BY
    1 
  "; 
  $sth = $dbh->prepare($sql) || &dbError("Prepare_L3546");
  $sth->execute || &dbErrorOK("execute helpdb db=$dbname");

  $backup_rowcnt = 0; 
  $sent_status = ""; 

  while($href = $sth->fetchrow_hashref)
  {
    $backup_rowcnt++;

    $href_line = ""; 
    foreach (sort keys %{$href} )
    {
      $href_colname = $_;   
      $href_value   = $href->{$href_colname};   
      $href_line   .= "$href_colname=$href_value ";
    }
    &msg("  Found BACKUP process - $href_line");

    $last_process_age_hours = $href->{Start_Age_Hours}; 
    &msg("    Age of this BACKUP process (in hours): $last_process_age_hours"); 
  }

  if ($backup_rowcnt == 0)
  {
    $last_process_age_hours = -1; 
    $href_line = "(No backups running)"; 
  }
  
  &msg("Results - SQL process details of oldest backup process: $href_line"); 
  print "$P LRBACKUP_AGE_DET: $href_line\n"; 
  &msg("Results - Age (hours) of oldest backup process: $last_process_age_hours"); 
  print "$P LRBACKUP_AGE_HRS: $last_process_age_hours\n"; 

  &msg("checkLRBackup - Done - Found $backup_rowcnt total BACKUPs running");
}

# ----------------------------------------------------------------------
sub checkSuspectPages
# ----------------------------------------------------------------------
{
  &msg("checkSuspectPages - Looking for rows in MSDB..SUSPECT_PAGES");

  if ($vershort < 90) 
  {
    &msg("This only works for SQL 2005+ - Skipping"); 
    return;
  }
  
  $sql = "select
  database_id FROM msdb..suspect_pages WHERE (event_type = 1 OR event_type = 2 OR event_type = 3);
  "; 
  $sth = $dbh->prepare($sql) || &dbError("Prepare_L3546");
  $sth->execute || &dbErrorOK("execute helpdb db=$dbname");

  my $suspect_rowcnt = 0; 
  while($href = $sth->fetchrow_hashref)
  {
    $suspect_rowcnt++;
  }

  if ($suspect_rowcnt > 0) 
  {
    &msg("$suspect_rowcnt rows found - VERY BAD - Investigate immediately"); 
    print "$P SUSPECT_ROWCNT: $suspect_rowcnt\n"; 
  }
  else
  {
    &msg("No rows found - GOOD"); 
  }
  
  &msg("checkSuspectPages - Done - Found $suspect_rowcnt total SUSPECT_PAGES rows");
}

# ----------------------------------------------------------------------
sub checkSignalWait
# ----------------------------------------------------------------------
{
  &msg("checkSignalWait - A performance check");

  if ($vershort < 90) 
  {
    &msg("This only works for SQL 2005+ - Skipping"); 
    return;
  }
  
  $sql = "SELECT
            signal_wait_time_ms = SUM(signal_wait_time_ms) ,
	    '%signal waits' = CAST(100.0 * SUM(signal_wait_time_ms)
	    / SUM(wait_time_ms) AS NUMERIC(20, 2)) ,
	    resource_wait_time_ms = SUM(wait_time_ms - signal_wait_time_ms) ,
	    '%resource waits' = CAST(100.0 * SUM(wait_time_ms
						 - signal_wait_time_ms)
	    / SUM(wait_time_ms) AS NUMERIC(20, 2))
    FROM    master.sys.dm_os_wait_stats
    WHERE   wait_type NOT IN ( 'BROKER_EVENTHANDLER', 'BROKER_RECEIVE_WAITFOR',
			       'BROKER_TASK_STOP', 'BROKER_TO_FLUSH',
			       'BROKER_TRANSMITTER', 'CHECKPOINT_QUEUE', 'CHKPT',
			       'CLR_AUTO_EVENT', 'CLR_MANUAL_EVENT',
			       'CLR_SEMAPHORE', 'DISPATCHER_QUEUE_SEMAPHORE',
			       'FT_IFTS_SCHEDULER_IDLE_WAIT', 'FT_IFTSHC_MUTEX',
			       'KSOURCE_WAKEUP', 'LAZYWRITER_SLEEP',
			       'LOGMGR_QUEUE', 'ONDEMAND_TASK_QUEUE',
			       'REQUEST_FOR_DEADLOCK_SEARCH', 'RESOURCE_QUEUE',
			       'SERVER_IDLE_CHECK', 'SLEEP', 'SLEEP_BPOOL_FLUSH',
			       'SLEEP_DBSTARTUP', 'SLEEP_DCOMSTARTUP',
			       'SLEEP_MSDBSTARTUP', 'SLEEP_SYSTEMTASK',
			       'SLEEP_TASK', 'SLEEP_TEMPDBSTARTUP',
			       'SNI_HTTP_ACCEPT', 'SQLTRACE_BUFFER_FLUSH',
			       'SQLTRACE_INCREMENTAL_FLUSH_SLEEP', 'Total',
			       'TRACEWRITE,WAIT_FOR_RESULTS', 'WAITFOR',
			       'WAITFOR_TASKSHUTDOWN', 'XE_DISPATCHER_JOIN',
			       'XE_DISPATCHER_WAIT', 'XE_TIMER_EVENT' )
  "; 
  $sth = $dbh->prepare($sql) || &dbError("Prepare_L4522");
  $sth->execute              || &dbErrorOK("Execute_L4523 helpdb db=$dbname");

  $rowcnt = 0; 
  while (@Data = $sth->fetchrow_array)
  {
    $rowcnt++; 
    $signal_wait_time_ms   = shift(@Data); 
    $signal_wait_pct       = shift(@Data); 
    $resource_wait_time_ms = shift(@Data); 
    $resource_wait_pct     = shift(@Data); 
  }
  &msg("  Rows retreived: $rowcnt"); 

  &msg("  Results:"); 
  &msg("    signal_wait_time_ms=  $signal_wait_time_ms"); 
  &msg("    signal_wait_pct=      $signal_wait_pct"); 
  &msg("    resource_wait_time_ms=$resource_wait_time_ms"); 
  &msg("    resource_wait_pct=    $resource_wait_pct"); 

  &msg("PHONEHOME: checkSignalWait - Signal+Resource wait != 100") if ($signal_wait_pct + $resource_wait_pct) != 100; 

  if ($resource_wait_pct > 0) 
  {
    print "$P SIGNAL_Wait_Pct: $signal_wait_pct\n"; 
  }
  else
  {
    &msg("No SIGNAL WAIT response"); 
  }
  
  &msg("checkSignalWait - Done");
}

# ----------------------------------------------------------------------
sub checkBackupDBFull
# ----------------------------------------------------------------------
{
  &msg("checkBackupDBFull - Looking for DBs NOT in SIMPLE logging and with NO TLOG backups");

  if ($vershort < 90) 
  {
    &msg("This only works for SQL 2005+ - Skipping"); 
    return;
  }
  
  $sql = "SELECT  name
FROM    master.sys.databases
WHERE   recovery_model_desc != 'SIMPLE'
        AND is_read_only = 0
        AND name NOT IN (
        SELECT  DISTINCT
                database_name
        FROM    [msdb].[dbo].[backupset]
        WHERE   type = 'L'
                AND backup_start_date > DATEADD(dd, -30, GETDATE()) )
  "; 
  $sth = $dbh->prepare($sql) || &dbError("Prepare_L3546");
  $sth->execute || &dbErrorOK("execute helpdb db=$dbname");

  my $db_list  = ""; 
  my $db_count = 0; 
  while (@Data = $sth->fetchrow_array)
  {
    $db_count++;
    $db_list .= "$Data[0],"; 
  }

  if ($db_count > 0) 
  {
    $db_list =~ s/\s//g;
    $db_list =~ s/,$//g;
    &msg("$db_count DBs found - db_list=$db_list"); 
    print "$P DBS_FULL_LOGGING_NO_TLOG_BACKUP: $db_count $db_list\n"; 
  }
  else
  {
    &msg("No rows found - GOOD"); 
  }
  
  &msg("checkBackupDBFull - Done - Found $db_count DBs in exception status");
}

# ----------------------------------------------------------------------
sub checkVLF
# ----------------------------------------------------------------------
{
  &msg("checkVLF - Counting VLFs for each DB");

  $sql = "if OBJECT_ID('tempdb..#dbamon_vlf_stage') > 0 BEGIN
      drop table #dbamon_vlf_stage
    END"; 
  $dbh->do("$sql") || &dbError("Drop_If_L4697");
  $sql = "if OBJECT_ID('tempdb..#dbamon_vlf_results') > 0 BEGIN
      drop table #dbamon_vlf_results
    END"; 
  $dbh->do("$sql") || &dbError("Drop_If_L4701");
  $sql = "CREATE TABLE #dbamon_vlf_stage
	(
	  FileID INT ,
	  FileSize BIGINT ,
	  StartOffset BIGINT ,
	  FSeqNo BIGINT ,
	  [Status] BIGINT ,
	  Parity BIGINT ,
	  CreateLSN NUMERIC(38)
	) ;"; 
  $dbh->do("$sql") || &dbError("Create_4712");
  $sql = "CREATE TABLE #dbamon_vlf_results
	(
	  Database_Name SYSNAME ,
	  VLF_count INT
	) ;"; 
  $dbh->do("$sql") || &dbError("Create_4718");
  $sql = "EXEC sp_msforeachdb N'Use [?];
		Insert Into #dbamon_vlf_stage
		Exec master..sp_executeSQL N''DBCC LogInfo([?]) with no_infomsgs'';
     
		Insert Into #dbamon_vlf_results
		Select DB_Name(), Count(*)
		From #dbamon_vlf_stage;
     
		Truncate Table #dbamon_vlf_stage;'"; 
  $dbh->do("$sql") || &dbError("Msforeachdb_4728");

  $sql = "
    SELECT  database_name, vlf_count
    FROM    #dbamon_vlf_results
    ORDER BY VLF_count DESC ;"; 
  $sth = $dbh->prepare($sql) || &dbError("Prepare_L4734");
  $sth->execute || &dbErrorOK("Select_4735");

  my $vlf_total = 0; 
  my $vlf_dbname = ""; 
  my $vlf_count = 0; 

  while (@Data = $sth->fetchrow_array)
  {
    $vlf_dbname = $Data[0]; 
    $vlf_count  = $Data[1]; 

    $vlf_total += $vlf_count; 
    $vlf_dbname =~ s/\s/[BLANK]/g;
    $vlf_dbname =~ s/\-/[HYPHEN]/g;

    print "$P VLF_COUNT: $vlf_dbname $vlf_count\n"; 
  }

  $sql = " DROP TABLE #dbamon_vlf_stage ;";  
  $dbh->do("$sql") || &dbError("Drop_If_L1697");
  $sql = " DROP TABLE #dbamon_vlf_results ;"; 
  $dbh->do("$sql") || &dbError("Drop_If_L1697");

  &msg("checkVLF - Done - Found $vlf_total Total VLFs");
}

# ----------------------------------------------------------------------
sub checkJobs
# ----------------------------------------------------------------------
{
  &msg("checkJobs - Listing all MSDB jobs");

  $sth = $dbh->prepare("exec msdb..sp_help_job") || &dbError("Prepare_L3193");
  $sth->execute || &dbErrorOK("execute helpdb db=$dbname");

  $job_cnt = 0; 

  #while($href = $sth->fetchrow_hashref)
  #{
  #  $job_name     = $href->{name}; 
  #  $job_enabled  = $href->{enabled}; 
  #  $job_category = $href->{category}; 
  #  $job_created  = $href->{date_created}; 
  #  $job_modified = $href->{date_modified}; 
  #  $job_status   = $href->{current_execution_status}; 
  #  $job_lr_date    = $href->{last_run_date}; 
  #  $job_lr_time    = $href->{last_run_time}; 
  #  $job_lr_outcome = $href->{last_run_outcome}; 
  #  ##&msg("DEBUG - job_name=$job_name"); 
  #  print "DDP:<|>$P<|>JOBS:<|>$job_name<|>$job_status<|>$job_enabled<|>$job_category<|>$job_created<|>$job_modified<|>$job_lr_date<|>$job_lr_time<|>$job_lr_outcome<|>\n";
  #  $job_cnt++; 
  #}

  while($href = $sth->fetchrow_hashref)
  {
    $job_cudata = ""; 
    foreach (sort keys %{$href} )
    {
      $job_colname = $_;   
      $job_value   = $href->{$job_colname};   
      $job_value   =~ s/\n/ /g;
      $job_cudata .= "$job_colname='$job_value'<|>"
    }
    print "DDP:<|>$P<|>JOBS:<|>$job_cudata\n";
    $job_cnt++; 
  }

  &msg("checkJobs - Done - Found $job_cnt jobs");
}

# ----------------------------------------------------------------------
sub saveConfigSQL
# ----------------------------------------------------------------------
{
  $scs_db  = $_[0]; 
  $scs_sql = $_[1]; 
  $scs_width = $_[2]; 
  $scs_width = 28 if ! $scs_width;
  &msg("saveConfigSQL - Starting - $scs_db=$scs_db scs_sql=$scs_sql scs_width=$scs_width");

  $dbh->do("use \[$scs_db\]")           || &dbError("scs_Use_Master_1");
  $sth = $dbh->prepare("$scs_sql ") || &dbError("scs_Prepare");
  $sth->execute                     || &dbError("scs_Execute_Select");

  push @saveconfig_out, " ";
  push @saveconfig_out, "================================================================================ ";
  push @saveconfig_out, "Running: $scs_sql - Against DB: $scs_db";
  push @saveconfig_out, "================================================================================ ";
  push @saveconfig_out, " ";
  $scs_heading = 0; 

  do 
  {
    while($href = $sth->fetchrow_hashref)
    {
      if (! $scs_heading)
      {
	$scs_line = ""; 
	foreach (sort keys %{$href} )
	{
	  $scs_colname = sprintf("%-${scs_width}s",$_);   
	  $scs_line    .= uc "$scs_colname "; 
	}
	push @saveconfig_out, "$scs_line";
	$scs_heading = 1; 
      }

      $scs_line = ""; 
      foreach (sort keys %{$href} )
      {
	$scs_colname = $_;   
	$scs_value   = sprintf("%-${scs_width}s",$href->{$scs_colname});   
	$scs_line .= "$scs_value "; 
      }
      push @saveconfig_out, "$scs_line";
    }
  }
  while($sth->{odbc_more_results});
  
  &msg("saveConfigSQL - Done");
}

# ----------------------------------------------------------------------
sub saveConfig
# ----------------------------------------------------------------------
{
  &msg("saveConfig - Listing instance and DB info");
  $sc_tsfile = "dbamon_mssql_saveconfig_$in_instance.txt";
  $sc_seconds_ago_t = 43200; 

  my $bookmark_from_file = 0; 

  if ($bookmark_from_file)
  {
    #---- (FILE Start) Read TS file
    open(TS, "$sc_tsfile") || print "$P | Can't open input $sc_tsfile - $!\n";;
    while (<TS>)
    {
      @words = split;
      $tsts   = $words[0];
      $tstime = $words[1];
    }
    close(TS);
    $tstime_seconds_ago = time - $tstime; 
    &msg("From $sc_tsfile - tstime=$tstime tstime_seconds_ago=$tstime_seconds_ago");
    if ($tstime > 0 && $tstime_seconds_ago < $sc_seconds_ago_t )
    {
      &msg("Since it has only been $tstime_seconds_ago seconds since this was last run, skipping (Threshold: $sc_seconds_ago_t seconds)"); 
      return;
    }
    #---- (FILE End) Read TS file
  }

  $tstime_seconds_ago = time - $dcf_data{'SaveConfigTS_Num'}; 
  &msg("DCF - dcf_data{'SaveConfigTS_Char'}=$dcf_data{'SaveConfigTS_Char'} - tstime_seconds_ago=$tstime_seconds_ago");
  &msg("saveConfing Was last run $tstime_seconds_ago seconds ago ($dcf_data{'SaveConfigTS_Char'})");
  if ($dcf_data{'SaveConfigTS_Num'} > 0 && $tstime_seconds_ago < $sc_seconds_ago_t )
  {
    &msg("Since it has only been $tstime_seconds_ago seconds since this was last run, skipping this execution (Threshold: $sc_seconds_ago_t seconds)"); 
    return;
  }

  #---- Steps to save configuration
  &connectDB if ! $connected_to_db; 
  undef @saveconfig_out;

  #---- Heading
  $date = &get_date("Y/M/D-H:M:S");
  push @saveconfig_out, " ";
  push @saveconfig_out, "================================================================================ ";
  push @saveconfig_out, "SaveConfig Report - Created on $host at $date ";
  push @saveconfig_out, "================================================================================ ";
  push @saveconfig_out, " ";

  #---- Get list of all DBs 
  $dbh->do("use master") || &dbError("Use_Master");
  undef @sc_dblist;
  $sql = "select name from sysdatabases order by 1; ";
  $sth = $dbh->prepare("$sql") || &dbError("Prepare");
  $sth->execute                || &dbError("Execute_Select");
  while (@Data = $sth->fetchrow_array)
  {
    $dbname = $Data[0];
    push(@sc_dblist,$dbname);
    ++$numdbs;
  }
  foreach $this_db (@sc_dblist)
  {
    &msg("SC db=$this_db"); 

    #---- This was for blanks and hyphens in the DB name
    #if ($zthis_db =~ /\s/ || $zthis_db =~ /\-/)
    #{
    #  &msg("Not processing DB with funky name - $this_db"); 
    #}
    #else
    #{
    #  &saveConfigSQL($this_db,"sp_helpdb \@dbname='$this_db'"); 
    #  &saveConfigSQL($this_db,"sp_dboption '$this_db'"); 
    #}

    &saveConfigSQL($this_db,"sp_helpdb \@dbname='$this_db'"); 
    &saveConfigSQL($this_db,"sp_dboption '$this_db'"); 
  }

  &saveConfigSQL("master","sp_configure"); 
  &saveConfigSQL("master","sp_helpsort"); 
  &saveConfigSQL("master","sp_helpdevice"); 
  &saveConfigSQL("master","sp_helpdb"); 

  foreach (@saveconfig_out)
  {
    print "$P Stack: SaveConfig: $_ \n"; 
  }

  #---- Save timestamp of this sample
  $date = &get_date("Y/M/D-H:M:S");
  $timestamp = time;

  if ($bookmark_from_file)
  {
    #---- Old - Save timestamp
    &msg("Saving timestamps: $save_rdate $save_rtime to $sc_tsfile");
    open(TS, "> $sc_tsfile")    || die "Error: Can't open output file: $sc_tsfile\n";
    print TS "$date $timestamp\n";
    close(TS);
    &msg("Contents of sc_tsfile=$sc_tsfile:");
    system("type $sc_tsfile"); 
  }

  #---- New - Save timestamp
  &msg("Saving timestamps: $save_rdate $save_rtime to DCF");
  $dcf_data{'SaveConfigTS_Char'} = "$date";
  $dcf_data{'SaveConfigTS_Num'} = "$timestamp";

  &msg("saveConfig - Done");
}

#--------------------------------------------------------------------------
#
# Main
#
#--------------------------------------------------------------------------

&msg("Starting - host=$host osname=$osname userid=$userid instance=($in_instance) domain=$domain pshort=$pshort");

#---- Initial housekeeping
&init;
&DCFRead;

#---- Find SQL instance, version, system stuff 
&getRegInfo;

#---- Check to see if Oracle is up
if ($verb eq "up") 
{
  ##$subrc += &SQLPerf;  #---- Move to within the "up" sub, since the DB may not be running
  $subrc += &checkDoubleTake(); #---- if $host eq "wcosdba2" || $host eq "wgttdrp";
  $subrc += &up();
  $subrc += &oldPerl();
  $subrc += &checkTLog();
  $subrc += &checkFilegroups();
  $subrc += &checkJobs();
  $subrc += &checkDBMail();
  $subrc += &checkLRBackup; 
  $subrc += &checkSuspectPages; 
  $subrc += &checkSignalWait; 
  $subrc += &checkVLF; 
  $subrc += &saveConfig();
}

#---- Check the event log
if ($verb eq "checklog") 
{
  $subrc =  &checklog();
  $subrc += &checkAgentLog unless $sql_edition =~ /express/i; 
}

#---- Check TLOGs
if ($verb eq "checktlog") 
{
  $subrc = &checkTLog();
}

#---- Backup age
if ($verb eq "bkup") 
{
  $subrc =  &bkup();
  $subrc += &checkBackupDBFull;
  #---- Not used any more
  # $subrc = &backupJobs();
}

#---- Backup rerun
if ($verb eq "bkrun") 
{
  $subrc = &bkrun();
}

#---- Backup logs - run from menu
if ($verb eq "bkbl") 
{
  $subrc = &bkbl();
}

#---- Backup history
if ($verb eq "bkhi") 
{
  $subrc = &bkhi();
}

#---- Test - Backup history
if ($verb eq "test1") 
{
  ##$subrc = &backupJobs();
  ##$subrc += &SysPerf; 
  &connectDB;
  #$subrc += &checkLRBackup; 
  #$dbname = "bbtest3 bdprms"; 
  #$subrc += &getDBAttrs; 
  #$subrc += &checkSignalWait; 
  $subrc += &checkVLF; 
}

#---- Test - Old Perl ProcessesBackup history
if ($verb eq "test2") 
{
  ##$subrc = &oldPerl();
  ##$subrc = &checkFilegroups();
  #$subrc = &saveConfig();
  #$subrc = &driveCheck();
  #$subrc = &driveCheckCMD();

  $subrc += &SysPerf; 
  $subrc += &SysPerfBulk; 
  $subrc += &SysPerfCompare; 
  $subrc += &SysPerfSpecial; 
}

#---- Terminate
$etime = time;
$elapsed = $etime - $stime;

&DCFWrite;

&msg("Done - elapsed: $elapsed - subrc: $subrc");
&msg("END_OF_PROBE");
close(PLOG); 
exit 0;
